"""
Adapted from COMBO
Author: Mateusz Klimaszewski
"""
from typing import Tuple, Dict, Optional, Union, List

import numpy as np
import torch
import torch.nn.functional as F
from overrides import overrides

from combo import data
from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.nn import chu_liu_edmonds, base
from combo.nn.utils import masked_cross_entropy
from combo.predictors.predictor import Predictor


@Registry.register('head_prediction')
class HeadPredictionModel(Predictor):
    """Head prediction model."""

    @register_arguments
    def __init__(self,
                 head_projection_layer: base.Linear,
                 dependency_projection_layer: base.Linear,
                 cycle_loss_n: int = 0):
        super().__init__()
        self.head_projection_layer = head_projection_layer
        self.dependency_projection_layer = dependency_projection_layer
        self.cycle_loss_n = cycle_loss_n

    @overrides
    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        if mask is None:
            mask = x.new_ones(x.size()[-1])

        head_arc_emb = self.head_projection_layer(x)
        dep_arc_emb = self.dependency_projection_layer(x)
        x = dep_arc_emb.bmm(head_arc_emb.transpose(2, 1))

        if self.training:
            pred = x.argmax(-1)
        else:
            pred = []
            # Adding non existing in mask ROOT to lengths
            lengths = mask.data.sum(dim=1).long().cpu().numpy() + 1
            for idx, length in enumerate(lengths):
                probs = x[idx, :].softmax(dim=-1).cpu().numpy()

                # We do not want any word to be parent of the root node (ROOT, 0).
                # Also setting it to -1 instead of 0 fixes edge case where softmax made all
                # but ROOT prediction to EXACTLY 0.0 and it might cause in many ROOT -> word edges)
                probs[:, 0] = -1
                heads, _ = chu_liu_edmonds.decode_mst(probs.T, length=length, has_labels=False)
                heads[0] = 0
                pred.append(heads)
            pred = torch.from_numpy(np.stack(pred)).to(x.device)

        output = {
            "prediction": pred[:, 1:],
            "probability": x
        }

        if labels is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            output["loss"], output["cycle_loss"] = self._loss(x, labels, mask, sample_weights)

        return output

    def _cycle_loss(self, pred: torch.Tensor):
        BATCH_SIZE, _, _ = pred.size()
        loss = pred.new_zeros(BATCH_SIZE)
        # Index from 1: as using non __ROOT__ tokens
        pred = pred.softmax(-1)[:, 1:, 1:]
        x = pred
        for i in range(self.cycle_loss_n):
            loss += self._batch_trace(x)

            # Don't multiple on last iteration
            if i < self.cycle_loss_n - 1:
                x = x.bmm(pred)

        return loss

    @staticmethod
    def _batch_trace(x: torch.Tensor) -> torch.Tensor:
        assert len(x.size()) == 3
        BATCH_SIZE, N, M = x.size()
        assert N == M
        identity = x.new_tensor(torch.eye(N))
        identity = identity.reshape((1, N, N))
        batch_identity = identity.repeat(BATCH_SIZE, 1, 1)
        return (x * batch_identity).sum((-1, -2))

    def _loss(self, pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        BATCH_SIZE, N, M = pred.size()
        assert N == M
        SENTENCE_LENGTH = N

        valid_positions = mask.sum()

        result = []
        # Ignore first pred dimension as it is ROOT token prediction
        for i in range(SENTENCE_LENGTH - 1):
            pred_i = pred[:, i + 1, :].reshape(BATCH_SIZE, SENTENCE_LENGTH)
            true_i = true[:, i].reshape(-1)
            mask_i = mask[:, i]
            cross_entropy_loss = masked_cross_entropy(pred_i, true_i, mask_i)
            result.append(cross_entropy_loss)
        cycle_loss = self._cycle_loss(pred)
        loss = torch.stack(result).transpose(1, 0) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions + cycle_loss.mean(), cycle_loss.mean()


@Registry.register('combo_dependency_parsing_from_vocab')
class DependencyRelationModel(Predictor):
    """Dependency relation parsing model."""

    @register_arguments
    def __init__(self,
                 vocabulary: data.Vocabulary,
                 vocab_namespace: str,
                 head_predictor: HeadPredictionModel,
                 head_projection_layer: base.Linear,
                 dependency_projection_layer: base.Linear
                 ):
        """Creates parser combining model configuration and vocabulary data."""
        super().__init__()
        assert vocab_namespace in vocabulary.get_namespaces()
        relation_prediction_layer = base.Linear(
            in_features=head_projection_layer.get_output_dim() + dependency_projection_layer.get_output_dim(),
            out_features=vocabulary.get_vocab_size(vocab_namespace)
        )

        self.root_idx = vocabulary.get_token_index("root", vocab_namespace)
        self.head_predictor = head_predictor
        self.head_projection_layer = head_projection_layer
        self.dependency_projection_layer = dependency_projection_layer
        self.relation_prediction_layer = relation_prediction_layer

    @overrides
    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        device = x.device
        if mask is not None:
            mask = mask[:, 1:]
        relations_labels, head_labels = None, None
        if labels is not None and labels[0] is not None:
            relations_labels, head_labels = labels
            if mask is None:
                mask = head_labels.new_ones(head_labels.size())

        head_output = self.head_predictor(x, mask, head_labels, sample_weights)
        head_pred = head_output["probability"]
        head_pred_soft = F.softmax(head_pred, dim=-1)

        head_rel_emb = self.head_projection_layer(x)

        dep_rel_emb = self.dependency_projection_layer(x)

        dep_rel_pred = head_pred_soft.bmm(head_rel_emb)
        dep_rel_pred = torch.cat((dep_rel_pred, dep_rel_emb), dim=-1)
        relation_prediction = self.relation_prediction_layer(dep_rel_pred)
        output = head_output
        output["embedding"] = dep_rel_pred
        output["deprel_label_distribution"] = F.softmax(relation_prediction[:, 1:, 1:], dim=-1)
        output["deprel_tree_distribution"] = head_pred_soft

        if self.training:
            output["prediction"] = (relation_prediction.argmax(-1)[:, 1:], head_output["prediction"])
        else:
            # Mask root label whenever head is not 0.
            relation_prediction_output = relation_prediction[:, 1:].clone()
            mask = (head_output["prediction"] == 0)
            vocab_size = relation_prediction_output.size(-1)
            root_idx = torch.tensor([self.root_idx], device=device)
            relation_prediction_output[mask] = (relation_prediction_output
                                                .masked_select(mask.unsqueeze(-1))
                                                .reshape(-1, vocab_size)
                                                .index_fill(-1, root_idx, 10e10))
            relation_prediction_output[~mask] = (relation_prediction_output
                                                 .masked_select(~(mask.unsqueeze(-1)))
                                                 .reshape(-1, vocab_size)
                                                 .index_fill(-1, root_idx, -10e10))
            output["prediction"] = (relation_prediction_output.argmax(-1), head_output["prediction"])

        if labels is not None and labels[0] is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            loss = self._loss(relation_prediction[:, 1:], relations_labels, mask, sample_weights)
            output["loss"] = (loss, head_output["loss"])

        return output

    @staticmethod
    def _loss(pred: torch.Tensor,
              true: torch.Tensor,
              mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:

        valid_positions = mask.sum()

        BATCH_SIZE, _, DEPENDENCY_RELATIONS = pred.size()
        pred = pred.reshape(-1, DEPENDENCY_RELATIONS)
        true = true.reshape(-1)
        mask = mask.reshape(-1)
        loss = masked_cross_entropy(pred, true, mask)
        loss = loss.reshape(BATCH_SIZE, -1) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions



import unittest
import os

from combo.data import UniversalDependenciesDatasetReader


class UniversalDependenciesDatasetReaderTest(unittest.TestCase):
    def test_read_all_tokens(self):
        t = UniversalDependenciesDatasetReader()
        tokens = [token for token in t.read(os.path.join(os.path.dirname(__file__), 'tl_trg-ud-test.conllu'))]
        self.assertEqual(len(tokens), 128)

    def test_read_text(self):
        t = UniversalDependenciesDatasetReader()
        token = next(iter(t.read(os.path.join(os.path.dirname(__file__), 'tl_trg-ud-test.conllu'))))
        self.assertListEqual([t.text for t in token['sentence'].tokens],
                             ['Gumising', 'ang', 'bata', '.'])

    def test_read_deprel(self):
        t = UniversalDependenciesDatasetReader()
        token = next(iter(t.read(os.path.join(os.path.dirname(__file__), 'tl_trg-ud-test.conllu'))))
        self.assertListEqual(token['deprel'].labels,
                             ['root', 'case', 'nsubj', 'punct'])

    def test_read_upostag(self):
        t = UniversalDependenciesDatasetReader()
        token = next(iter(t.read(os.path.join(os.path.dirname(__file__), 'tl_trg-ud-test.conllu'))))
        self.assertListEqual(token['upostag'].labels,
                             ['VERB', 'ADP', 'NOUN', 'PUNCT'])




"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/tokenizers/spacy_tokenizer.py
"""

from typing import List, Optional

import spacy
from spacy.tokens import Doc

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.tokenizers.token import Token
from combo.data.tokenizers.tokenizer import Tokenizer
from combo.utils.spacy import get_spacy_model


@Registry.register('spacy_tokenizer')
class SpacyTokenizer(Tokenizer):
    """
    A `Tokenizer` that uses spaCy's tokenizer.  It's fast and reasonable - this is the
    recommended `Tokenizer`. By default it will return allennlp Tokens,
    which are small, efficient NamedTuples (and are serializable). If you want
    to keep the original spaCy tokens, pass keep_spacy_tokens=True.  Note that we leave one particular piece of
    post-processing for later: the decision of whether or not to lowercase the token.  This is for
    two reasons: (1) if you want to make two different casing decisions for whatever reason, you
    won't have to run the tokenizer twice, and more importantly (2) if you want to lowercase words
    for your word embedding, but retain capitalization in a character-level representation, we need
    to retain the capitalization here.
    Registered as a `Tokenizer` with name "spacy", which is currently the default.
    # Parameters
    language : `str`, optional, (default=`"en_core_web_sm"`)
        Spacy model name.
    pos_tags : `bool`, optional, (default=`False`)
        If `True`, performs POS tagging with spacy model on the tokens.
        Generally used in conjunction with :class:`~allennlp.data.token_indexers.pos_tag_indexer.PosTagIndexer`.
    parse : `bool`, optional, (default=`False`)
        If `True`, performs dependency parsing with spacy model on the tokens.
        Generally used in conjunction with :class:`~allennlp.data.token_indexers.pos_tag_indexer.DepLabelIndexer`.
    ner : `bool`, optional, (default=`False`)
        If `True`, performs dependency parsing with spacy model on the tokens.
        Generally used in conjunction with :class:`~allennlp.data.token_indexers.ner_tag_indexer.NerTagIndexer`.
    keep_spacy_tokens : `bool`, optional, (default=`False`)
        If `True`, will preserve spacy token objects, We copy spacy tokens into our own class by default instead
        because spacy Cython Tokens can't be pickled.
    split_on_spaces : `bool`, optional, (default=`False`)
        If `True`, will split by spaces without performing tokenization.
        Used when your data is already tokenized, but you want to perform pos, ner or parsing on the tokens.
    start_tokens : `Optional[List[str]]`, optional, (default=`None`)
        If given, these tokens will be added to the beginning of every string we tokenize.
    end_tokens : `Optional[List[str]]`, optional, (default=`None`)
        If given, these tokens will be added to the end of every string we tokenize.
    """

    @register_arguments
    def __init__(
            self,
            language: str = "en_core_web_sm",
            pos_tags: bool = True,
            parse: bool = False,
            ner: bool = False,
            keep_spacy_tokens: bool = False,
            split_on_spaces: bool = False,
            start_tokens: Optional[List[str]] = None,
            end_tokens: Optional[List[str]] = None,
    ) -> None:
        # Save these for use later in the _to_params method
        self._language = language
        self._pos_tags = pos_tags
        self._parse = parse
        self._ner = ner
        self._split_on_spaces = split_on_spaces

        self.spacy = get_spacy_model(self._language, self._pos_tags, self._parse, self._ner)

        if self._split_on_spaces:
            self.spacy.tokenizer = _WhitespaceSpacyTokenizer(self.spacy.vocab)

        self._keep_spacy_tokens = keep_spacy_tokens
        self._start_tokens = start_tokens or []
        # We reverse the tokens here because we're going to insert them with `insert(0)` later;
        # this makes sure they show up in the right order.
        self._start_tokens.reverse()
        self._is_version_3 = spacy.__version__ >= "3.0"
        self._end_tokens = end_tokens or []

    def _sanitize(self, tokens: List[spacy.tokens.Token]) -> List[Token]:
        """
        Converts spaCy tokens to allennlp tokens. Is a no-op if
        keep_spacy_tokens is True
        """
        # self.text = text
        # self.lemma = lemma
        # self.upostag = upostag
        # self.xpostag = xpostag
        # self.entity_type = entity_type
        # self.feats = feats
        # self.head = head
        # self.deprel = deprel
        # self.deps = deps
        # self.misc = misc
        # self.semrel = semrel

        # TODO: add morph from SpaCy?
        if not self._keep_spacy_tokens:
            tokens = [
                Token(
                    idx=(token.idx, token.idx + len(token.text)),
                    text=token.text,
                    lemma=token.lemma_,
                    upostag=token.pos_,
                    xpostag=token.tag_,
                    deprel=token.dep_,
                    entity_type=token.ent_type_
                )
                for token in tokens
            ]
        for start_token in self._start_tokens:
            tokens.insert(0, Token(idx=0, text=start_token))
        for end_token in self._end_tokens:
            tokens.append(Token(idx=-1, text=end_token))
        return tokens

    def batch_tokenize(self, texts: List[str]) -> List[List[Token]]:
        if self._is_version_3:
            return [
                self._sanitize(_remove_spaces(tokens))
                for tokens in self.spacy.pipe(texts, n_process=-1)
            ]
        else:
            return [
                self._sanitize(_remove_spaces(tokens))
                for tokens in self.spacy.pipe(texts, n_threads=-1)
            ]

    def tokenize(self, text: str) -> List[Token]:
        # This works because our Token class matches spacy's.
        return self._sanitize(_remove_spaces(self.spacy(text)))


class _WhitespaceSpacyTokenizer:
    """
    Spacy doesn't assume that text is tokenised. Sometimes this
    is annoying, like when you have gold data which is pre-tokenised,
    but Spacy's tokenisation doesn't match the gold. This can be used
    as follows:
    nlp = spacy.load("en_core_web_md")
    # hack to replace tokenizer with a whitespace tokenizer
    nlp.tokenizer = _WhitespaceSpacyTokenizer(nlp.vocabulary)
    ... use nlp("here is some text") as normal.
    """

    def __init__(self, vocab):
        self.vocab = vocab

    def __call__(self, text):
        words = text.split(" ")
        spaces = [True] * len(words)
        return Doc(self.vocab, words=words, spaces=spaces)


def _remove_spaces(tokens: List[spacy.tokens.Token]) -> List[spacy.tokens.Token]:
    return [token for token in tokens if not token.is_space]



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/nn/util.py
"""
from typing import Union, Dict, Optional, List, Any, NamedTuple

import torch

from combo.common.util import int_to_device
from combo.utils import ConfigurationError
import torch.nn.functional as F

StateDictType = Union[Dict[str, torch.Tensor], "OrderedDict[str, torch.Tensor]"]

def masked_cross_entropy(pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor) -> torch.Tensor:
    pred = pred + (mask.float().unsqueeze(-1) + 1e-45).log()
    return F.cross_entropy(pred, true, reduction="none") * mask

def tiny_value_of_dtype(dtype: torch.dtype):
    """
    Returns a moderately tiny value for a given PyTorch data type that is used to avoid numerical
    issues such as division by zero.
    This is different from `info_value_of_dtype(dtype).tiny` because it causes some NaN bugs.
    Only supports floating point dtypes.
    """
    if not dtype.is_floating_point:
        raise TypeError("Only supports floating point dtypes.")
    if dtype == torch.float or dtype == torch.double:
        return 1e-13
    elif dtype == torch.half:
        return 1e-4
    else:
        raise TypeError("Does not support dtype " + str(dtype))


def get_device_of(tensor: torch.Tensor) -> int:
    """
    Returns the device of the tensor.
    """
    if not tensor.is_cuda:
        return -1
    else:
        return tensor.get_device()


def combine_initial_dims(tensor: torch.Tensor) -> torch.Tensor:
    """
    Given a (possibly higher order) tensor of ids with shape
    (d1, ..., dn, sequence_length)
    Return a view that's (d1 * ... * dn, sequence_length).
    If original tensor is 1-d or 2-d, return it as is.
    """
    if tensor.dim() <= 2:
        return tensor
    else:
        return tensor.view(-1, tensor.size(-1))


def uncombine_initial_dims(tensor: torch.Tensor, original_size: torch.Size) -> torch.Tensor:
    """
    Given a tensor of embeddings with shape
    (d1 * ... * dn, sequence_length, embedding_dim)
    and the original shape
    (d1, ..., dn, sequence_length),
    return the reshaped tensor of embeddings with shape
    (d1, ..., dn, sequence_length, embedding_dim).
    If original size is 1-d or 2-d, return it as is.
    """
    if len(original_size) <= 2:
        return tensor
    else:
        view_args = list(original_size) + [tensor.size(-1)]
        return tensor.view(*view_args)


def get_range_vector(size: int, device: int) -> torch.Tensor:
    """
    Returns a range vector with the desired size, starting at 0. The CUDA implementation
    is meant to avoid copy data from CPU to GPU.
    """
    if device > -1:
        return torch.cuda.LongTensor(size, device=device).fill_(1).cumsum(0) - 1
    else:
        return torch.arange(0, size, dtype=torch.long)


def flatten_and_batch_shift_indices(indices: torch.Tensor, sequence_length: int) -> torch.Tensor:
    if torch.max(indices) >= sequence_length or torch.min(indices) < 0:
        raise ConfigurationError(
            f"All elements in indices should be in range (0, {sequence_length - 1})"
        )
    offsets = get_range_vector(indices.size(0), get_device_of(indices)) * sequence_length
    for _ in range(len(indices.size()) - 1):
        offsets = offsets.unsqueeze(1)

    # Shape: (batch_size, d_1, ..., d_n)
    offset_indices = indices + offsets

    # Shape: (batch_size * d_1 * ... * d_n)
    offset_indices = offset_indices.view(-1)
    return offset_indices


def batched_index_select(
        target: torch.Tensor,
        indices: torch.LongTensor,
        flattened_indices: Optional[torch.LongTensor] = None,
) -> torch.Tensor:
    if flattened_indices is None:
        # Shape: (batch_size * d_1 * ... * d_n)
        flattened_indices = flatten_and_batch_shift_indices(indices, target.size(1))

        # Shape: (batch_size * sequence_length, embedding_size)
    flattened_target = target.view(-1, target.size(-1))

    # Shape: (batch_size * d_1 * ... * d_n, embedding_size)
    flattened_selected = flattened_target.index_select(0, flattened_indices)
    selected_shape = list(indices.size()) + [target.size(-1)]
    # Shape: (batch_size, d_1, ..., d_n, embedding_size)
    selected_targets = flattened_selected.view(*selected_shape)
    return selected_targets


def move_to_device(obj, device: Union[torch.device, int]):
    """
    Given a structure (possibly) containing Tensors,
    move all the Tensors to the specified device (or do nothing, if they are already on
    the target device).
    """
    device = int_to_device(device)

    if isinstance(obj, torch.Tensor):
        # You may be wondering why we don't just always call `obj.to(device)` since that would
        # be a no-op anyway if `obj` is already on `device`. Well that works fine except
        # when PyTorch is not compiled with CUDA support, in which case even calling
        # `obj.to(torch.device("cpu"))` would result in an error.
        return obj if obj.device == device else obj.to(device=device)
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = move_to_device(value, device)
        return obj
    elif isinstance(obj, list):
        for i, item in enumerate(obj):
            obj[i] = move_to_device(item, device)
        return obj
    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        # This is the best way to detect a NamedTuple, it turns out.
        return obj.__class__(*(move_to_device(item, device) for item in obj))
    elif isinstance(obj, tuple):
        return tuple(move_to_device(item, device) for item in obj)
    else:
        return obj


def device_mapping(cuda_device: int):
    """
    In order to `torch.load()` a GPU-trained model onto a CPU (or specific GPU),
    you have to supply a `map_location` function. Call this with
    the desired `cuda_device` to get the function that `torch.load()` needs.
    """

    def inner_device_mapping(storage: torch.Storage, location) -> torch.Storage:
        if cuda_device >= 0:
            return storage.cuda(cuda_device)
        else:
            return storage

    return inner_device_mapping


def find_text_field_embedder(model: torch.nn.Module) -> torch.nn.Module:
    """
    Takes a `Model` and returns the `Module` that is a `TextFieldEmbedder`.  We return just the
    first one, as it's very rare to have more than one.  If there isn't a `TextFieldEmbedder` in the
    given `Model`, we raise a `ValueError`.
    """
    from combo.modules.text_field_embedders.text_field_embedder import TextFieldEmbedder

    for module in model.modules():
        if isinstance(module, TextFieldEmbedder):
            return module
    raise ValueError("Couldn't find TextFieldEmbedder!")


def get_lengths_from_binary_sequence_mask(mask: torch.BoolTensor) -> torch.LongTensor:
    """
    Compute sequence lengths for each batch element in a tensor using a
    binary mask.
    # Parameters
    mask : `torch.BoolTensor`, required.
        A 2D binary mask of shape (batch_size, sequence_length) to
        calculate the per-batch sequence lengths from.
    # Returns
    `torch.LongTensor`
        A torch.LongTensor of shape (batch_size,) representing the lengths
        of the sequences in the batch.
    """
    return mask.sum(-1)


def sort_batch_by_length(tensor: torch.Tensor, sequence_lengths: torch.Tensor):
    """
    Sort a batch first tensor by some specified lengths.
    # Parameters
    tensor : `torch.FloatTensor`, required.
        A batch first Pytorch tensor.
    sequence_lengths : `torch.LongTensor`, required.
        A tensor representing the lengths of some dimension of the tensor which
        we want to sort by.
    # Returns
    sorted_tensor : `torch.FloatTensor`
        The original tensor sorted along the batch dimension with respect to sequence_lengths.
    sorted_sequence_lengths : `torch.LongTensor`
        The original sequence_lengths sorted by decreasing size.
    restoration_indices : `torch.LongTensor`
        Indices into the sorted_tensor such that
        `sorted_tensor.index_select(0, restoration_indices) == original_tensor`
    permutation_index : `torch.LongTensor`
        The indices used to sort the tensor. This is useful if you want to sort many
        tensors using the same ordering.
    """

    if not isinstance(tensor, torch.Tensor) or not isinstance(sequence_lengths, torch.Tensor):
        raise ConfigurationError("Both the tensor and sequence lengths must be torch.Tensors.")

    sorted_sequence_lengths, permutation_index = sequence_lengths.sort(0, descending=True)
    sorted_tensor = tensor.index_select(0, permutation_index)

    index_range = torch.arange(0, len(sequence_lengths), device=sequence_lengths.device)
    # This is the equivalent of zipping with index, sorting by the original
    # sequence lengths and returning the now sorted indices.
    _, reverse_mapping = permutation_index.sort(0, descending=False)
    restoration_indices = index_range.index_select(0, reverse_mapping)
    return sorted_tensor, sorted_sequence_lengths, restoration_indices, permutation_index


def get_text_field_mask(
        text_field_tensors: Dict[str, Dict[str, torch.Tensor]],
        num_wrapping_dims: int = 0,
        padding_id: int = 0,
) -> torch.BoolTensor:
    """
    Takes the dictionary of tensors produced by a `TextField` and returns a mask
    with 0 where the tokens are padding, and 1 otherwise. `padding_id` specifies the idx of padding tokens.
    We also handle `TextFields` wrapped by an arbitrary number of `ListFields`, where the number of wrapping
    `ListFields` is given by `num_wrapping_dims`.
    If `num_wrapping_dims == 0`, the returned mask has shape `(batch_size, num_tokens)`.
    If `num_wrapping_dims > 0` then the returned mask has `num_wrapping_dims` extra
    dimensions, so the shape will be `(batch_size, ..., num_tokens)`.
    There could be several entries in the tensor dictionary with different shapes (e.g., one for
    word ids, one for character ids).  In order to get a token mask, we use the tensor in
    the dictionary with the lowest number of dimensions.  After subtracting `num_wrapping_dims`,
    if this tensor has two dimensions we assume it has shape `(batch_size, ..., num_tokens)`,
    and use it for the mask.  If instead it has three dimensions, we assume it has shape
    `(batch_size, ..., num_tokens, num_features)`, and sum over the last dimension to produce
    the mask.  Most frequently this will be a character idx tensor, but it could also be a
    featurized representation of each token, etc.
    If the input `text_field_tensors` contains the "mask" key, this is returned instead of inferring the mask.
    """
    masks = []
    for indexer_name, indexer_tensors in text_field_tensors.items():
        if "mask" in indexer_tensors:
            masks.append(indexer_tensors["mask"].bool())
    if len(masks) == 1:
        return masks[0]
    elif len(masks) > 1:
        # TODO(mattg): My guess is this will basically never happen, so I'm not writing logic to
        # handle it.  Should be straightforward to handle, though.  If you see this error in
        # practice, open an issue on github.
        raise ValueError("found two mask outputs; not sure which to use!")

    tensor_dims = [
        (tensor.dim(), tensor)
        for indexer_output in text_field_tensors.values()
        for tensor in indexer_output.values()
    ]
    tensor_dims.sort(key=lambda x: x[0])

    smallest_dim = tensor_dims[0][0] - num_wrapping_dims
    if smallest_dim == 2:
        token_tensor = tensor_dims[0][1]
        return token_tensor != padding_id
    elif smallest_dim == 3:
        character_tensor = tensor_dims[0][1]
        return (character_tensor != padding_id).any(dim=-1)
    else:
        raise ValueError("Expected a tensor with dimension 2 or 3, found {}".format(smallest_dim))


def get_dropout_mask(dropout_probability: float, tensor_for_masking: torch.Tensor):
    """
    Computes and returns an element-wise dropout mask for a given tensor, where
    each element in the mask is dropped out with probability dropout_probability.
    Note that the mask is NOT applied to the tensor - the tensor is passed to retain
    the correct CUDA tensor type for the mask.
    # Parameters
    dropout_probability : `float`, required.
        Probability of dropping a dimension of the input.
    tensor_for_masking : `torch.Tensor`, required.
    # Returns
    `torch.FloatTensor`
        A torch.FloatTensor consisting of the binary mask scaled by 1/ (1 - dropout_probability).
        This scaling ensures expected values and variances of the output of applying this mask
        and the original tensor are the same.
    """
    binary_mask = (torch.rand(tensor_for_masking.size()) > dropout_probability).to(
        tensor_for_masking.device
    )
    # Scale mask by 1/keep_prob to preserve output statistics.
    dropout_mask = binary_mask.float().div(1.0 - dropout_probability)
    return dropout_mask


def find_embedding_layer(model: torch.nn.Module) -> torch.nn.Module:
    """
    Takes a model (typically an AllenNLP `Model`, but this works for any `torch.nn.Module`) and
    makes a best guess about which module is the embedding layer.  For typical AllenNLP models,
    this often is the `TextFieldEmbedder`, but if you're using a pre-trained contextualizer, we
    really want layer 0 of that contextualizer, not the output.  So there are a bunch of hacks in
    here for specific pre-trained contextualizers.
    """
    # We'll look for a few special cases in a first pass, then fall back to just finding a
    # TextFieldEmbedder in a second pass if we didn't find a special case.
    from transformers.models.gpt2.modeling_gpt2 import GPT2Model
    from transformers.models.bert.modeling_bert import BertEmbeddings
    from transformers.models.albert.modeling_albert import AlbertEmbeddings
    from transformers.models.roberta.modeling_roberta import RobertaEmbeddings

    for module in model.modules():
        if isinstance(module, BertEmbeddings):
            return module.word_embeddings
        if isinstance(module, RobertaEmbeddings):
            return module.word_embeddings
        if isinstance(module, AlbertEmbeddings):
            return module.word_embeddings
        if isinstance(module, GPT2Model):
            return module.wte

    return None

    # for module in model.modules():
    #     if isinstance(module, TextFieldEmbedder):
    #
    #         if isinstance(module, BasicTextFieldEmbedder):
    #             # We'll have a check for single Embedding cases, because we can be more efficient
    #             # in cases like this.  If this check fails, then for something like hotflip we need
    #             # to actually run the text field embedder and construct a vector for each token.
    #             if len(module._token_embedders) == 1:
    #                 embedder = list(module._token_embedders.values())[0]
    #                 if isinstance(embedder, Embedding):
    #                     if embedder._projection is None:
    #                         # If there's a projection inside the Embedding, then we need to return
    #                         # the whole TextFieldEmbedder, because there's more computation that
    #                         # needs to be done than just multiply by an embedding matrix.
    #                         return embedder
    #         return module
    raise RuntimeError("No embedding module found!")


def get_token_offsets_from_text_field_inputs(
        text_field_inputs: List[Any],
) -> Optional[torch.Tensor]:
    """
    Given a list of inputs to a TextFieldEmbedder, tries to find token offsets from those inputs, if
    there are any.  You will have token offsets if you are using a mismatched token embedder; if
    you're not, the return value from this function should be None.  This function is intended to be
    called from a `forward_hook` attached to a `TextFieldEmbedder`, so the inputs are formatted just
    as a list.
    It's possible in theory that you could have multiple offsets as inputs to a single call to a
    `TextFieldEmbedder`, but that's an extremely rare use case (I can't really imagine anyone
    wanting to do that).  In that case, we'll only return the first one.  If you need different
    behavior for your model, open an issue on github describing what you're doing.
    """
    for input_index, text_field_input in enumerate(text_field_inputs):
        if not isinstance(text_field_input, dict):
            continue
        for input_value in text_field_input.values():
            if not isinstance(input_value, dict):
                continue
            for embedder_arg_name, embedder_arg_value in input_value.items():
                if embedder_arg_name == "offsets":
                    return embedder_arg_value
    return None


def _check_incompatible_keys(
        module, missing_keys: List[str], unexpected_keys: List[str], strict: bool
):
    error_msgs: List[str] = []
    if missing_keys:
        error_msgs.append(
            "Missing key(s) in state_dict: {}".format(", ".join(f'"{k}"' for k in missing_keys))
        )
    if unexpected_keys:
        error_msgs.append(
            "Unexpected key(s) in state_dict: {}".format(
                ", ".join(f'"{k}"' for k in unexpected_keys)
            )
        )
    if error_msgs and strict:
        raise RuntimeError(
            "Error(s) in loading state_dict for {}:\n\t{}".format(
                module.__class__.__name__, "\n\t".join(error_msgs)
            )
        )


class _IncompatibleKeys(NamedTuple):
    missing_keys: List[str]
    unexpected_keys: List[str]

    def __repr__(self):
        if not self.missing_keys and not self.unexpected_keys:
            return "<All keys matched successfully>"
        return f"(missing_keys = {self.missing_keys}, unexpected_keys = {self.unexpected_keys})"


def batched_span_select(target: torch.Tensor, spans: torch.LongTensor) -> torch.Tensor:
    """
    The given `spans` of size `(batch_size, num_spans, 2)` indexes into the sequence
    dimension (dimension 2) of the target, which has size `(batch_size, sequence_length,
    embedding_size)`.

    This function returns segmented spans in the target with respect to the provided span indices.

    # Parameters

    target : `torch.Tensor`, required.
        A 3 dimensional tensor of shape (batch_size, sequence_length, embedding_size).
        This is the tensor to be indexed.
    indices : `torch.LongTensor`
        A 3 dimensional tensor of shape (batch_size, num_spans, 2) representing start and end
        indices (both inclusive) into the `sequence_length` dimension of the `target` tensor.

    # Returns

    span_embeddings : `torch.Tensor`
        A tensor with shape (batch_size, num_spans, max_batch_span_width, embedding_size]
        representing the embedded spans extracted from the batch flattened target tensor.
    span_mask: `torch.BoolTensor`
        A tensor with shape (batch_size, num_spans, max_batch_span_width) representing the mask on
        the returned span embeddings.
    """
    # both of shape (batch_size, num_spans, 1)
    span_starts, span_ends = spans.split(1, dim=-1)

    # shape (batch_size, num_spans, 1)
    # These span widths are off by 1, because the span ends are `inclusive`.
    span_widths = span_ends - span_starts

    # We need to know the maximum span width so we can
    # generate indices to extract the spans from the sequence tensor.
    # These indices will then get masked below, such that if the length
    # of a given span is smaller than the max, the rest of the values
    # are masked.
    max_batch_span_width = span_widths.max().item() + 1

    # Shape: (1, 1, max_batch_span_width)
    max_span_range_indices = get_range_vector(max_batch_span_width, get_device_of(target)).view(
        1, 1, -1
    )
    # Shape: (batch_size, num_spans, max_batch_span_width)
    # This is a broadcasted comparison - for each span we are considering,
    # we are creating a range vector of size max_span_width, but masking values
    # which are greater than the actual length of the span.
    #
    # We're using <= here (and for the mask below) because the span ends are
    # inclusive, so we want to include indices which are equal to span_widths rather
    # than using it as a non-inclusive upper bound.
    span_mask = max_span_range_indices <= span_widths
    raw_span_indices = span_starts + max_span_range_indices
    # We also don't want to include span indices which greater than the sequence_length,
    # which happens because some spans near the end of the sequence
    # have a start index + max_batch_span_width > sequence_length, so we add this to the mask here.
    span_mask = span_mask & (raw_span_indices < target.size(1)) & (0 <= raw_span_indices)
    span_indices = raw_span_indices * span_mask

    # Shape: (batch_size, num_spans, max_batch_span_width, embedding_dim)
    span_embeddings = batched_index_select(target, span_indices)

    return span_embeddings, span_mask



import torch
import torch.nn.functional as F


def masked_cross_entropy(pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor) -> torch.Tensor:
    pred = pred + (mask.float().unsqueeze(-1) + 1e-45).log()
    return F.cross_entropy(pred, true, reduction="none") * mask


"""
Adapted from AllenNLP
"""


def tiny_value_of_dtype(dtype: torch.dtype):
    """
    Returns a moderately tiny value for a given PyTorch data type that is used to avoid numerical
    issues such as division by zero.
    This is different from `info_value_of_dtype(dtype).tiny` because it causes some NaN bugs.
    Only supports floating point dtypes.
    """
    if not dtype.is_floating_point:
        raise TypeError("Only supports floating point dtypes.")
    if dtype == torch.float or dtype == torch.double:
        return 1e-13
    elif dtype == torch.half:
        return 1e-4
    else:
        raise TypeError("Does not support dtype " + str(dtype))



import unittest
import os

from combo.data.dataset_readers import TextClassificationJSONReader
from combo.data.fields import LabelField, TextField, ListField
from combo.data.tokenizers import SpacySentenceSplitter


class TextClassificationJSONReaderTest(unittest.TestCase):
    def test_read_two_tokens(self):
        reader = TextClassificationJSONReader()
        tokens = [token for token in reader.read(os.path.join(os.path.dirname(__file__), 'text_classification_json_reader.json'))]
        self.assertEqual(len(tokens), 2)

    def test_read_two_examples_fields_without_sentence_splitting(self):
        reader = TextClassificationJSONReader()
        tokens = [token for token in reader.read(os.path.join(os.path.dirname(__file__), 'text_classification_json_reader.json'))]
        self.assertEqual(len(tokens[0].fields.items()), 2)
        self.assertIsInstance(tokens[0].fields.get('label'), LabelField)
        self.assertEqual(tokens[0].fields.get('label').label, 'label1')
        self.assertEqual(len(tokens[1].fields.items()), 2)
        self.assertIsInstance(tokens[1].fields.get('label'), LabelField)
        self.assertEqual(tokens[1].fields.get('label').label, 'label2')

    def test_read_two_examples_tokens_without_sentence_splitting(self):
        reader = TextClassificationJSONReader()
        tokens = [token for token in reader.read(os.path.join(os.path.dirname(__file__), 'text_classification_json_reader.json'))]
        self.assertEqual(len(tokens[0].fields.items()), 2)
        self.assertIsInstance(tokens[0].fields.get('tokens'), TextField)
        self.assertEqual(len(tokens[0].fields.get('tokens').tokens), 13)
        self.assertEqual(len(tokens[1].fields.items()), 2)
        self.assertIsInstance(tokens[1].fields.get('tokens'), TextField)
        self.assertEqual(len(tokens[1].fields.get('tokens').tokens), 6)

    def test_read_two_examples_tokens_with_sentence_splitting(self):
        reader = TextClassificationJSONReader(sentence_segmenter=SpacySentenceSplitter())
        tokens = [token for token in reader.read(os.path.join(os.path.dirname(__file__), 'text_classification_json_reader.json'))]
        self.assertEqual(len(tokens[0].fields.items()), 2)
        self.assertIsInstance(tokens[0].fields.get('tokens'), ListField)
        self.assertEqual(len(tokens[0].fields.get('tokens').field_list), 2)
        self.assertEqual(len(tokens[1].fields.items()), 2)
        self.assertIsInstance(tokens[1].fields.get('tokens'), ListField)
        self.assertEqual(len(tokens[1].fields.get('tokens').field_list), 1)



"""
Adapted from COMBO
Author: Mateusz Klimaszewski
https://gitlab.clarin-pl.eu/syntactic-tools/combo/-/blob/master/combo/models/embeddings.py
"""
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary
from combo.models.dilated_cnn import DilatedCnnEncoder
from combo.modules.token_embedders import TokenEmbedder

from typing import Optional
import torch
import torch.nn as nn
from combo import modules


# @token_embedders.TokenEmbedder.register("char_embeddings")
@Registry.register("char_embeddings_token_embedder")
class CharacterBasedWordEmbedder(TokenEmbedder):
    """Character-based word embeddings."""

    @register_arguments
    def __init__(self,
                 embedding_dim: int,
                 vocabulary: Vocabulary,
                 dilated_cnn_encoder: DilatedCnnEncoder,
                 vocab_namespace: str = "token_characters"):
        assert vocab_namespace in vocabulary.get_namespaces()
        super().__init__()
        self.char_embed = nn.Embedding(
            num_embeddings=vocabulary.get_vocab_size(vocab_namespace),
            embedding_dim=embedding_dim,
        )
        self.dilated_cnn_encoder = modules.TimeDistributed(dilated_cnn_encoder)
        self.output_dim = embedding_dim

    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        if char_mask is None:
            char_mask = x.new_ones(x.size())

        x = self.char_embed(x)
        x = x * char_mask.unsqueeze(-1).float()
        x = self.dilated_cnn_encoder(x.transpose(2, 3))
        return torch.max(x, dim=-1)[0]

    @overrides
    def get_output_dim(self) -> int:
        return self.output_dim



from typing import List

import numpy as np

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.samplers import BatchSampler


@Registry.register('token_count_batch_sampler')
class TokenCountBatchSampler(BatchSampler):
    @register_arguments
    def __init__(self, dataset, word_batch_size: int = 2500, shuffle_dataset: bool = True):
        self._index = 0
        self.shuffle_dataset = shuffle_dataset
        self.batch_dataset = self._batchify(dataset, word_batch_size)
        if shuffle_dataset:
            self._shuffle()

    @staticmethod
    def _batchify(dataset, word_batch_size) -> List[List[int]]:
        dataset = list(dataset)
        batches = []
        batch = []
        words_count = 0
        lengths = [len(instance.fields["sentence"].tokens) for instance in dataset]
        argsorted_lengths = np.argsort(lengths)
        for idx in argsorted_lengths:
            words_count += lengths[idx]
            batch.append(idx)
            if words_count > word_batch_size:
                batches.append(batch)
                words_count = 0
                batch = []
        return batches

    def __iter__(self):
        return self

    def __next__(self):
        if self._index >= len(self.batch_dataset):
            if self.shuffle_dataset:
                self._index = 0
                self._shuffle()
            raise StopIteration()

        batch = self.batch_dataset[self._index]
        self._index += 1
        return batch

    def _shuffle(self):
        indices = np.random.permutation(range(len(self.batch_dataset)))
        self.batch_dataset = np.array(self.batch_dataset)[indices].tolist()

    def __len__(self):
        return len(self.batch_dataset)



from .vocabulary import Vocabulary
from .samplers import TokenCountBatchSampler
from .instance import Instance
from .token_indexers import (SingleIdTokenIndexer, TokenIndexer, TokenFeatsIndexer)
from .tokenizers import (Tokenizer, Token, CharacterTokenizer, PretrainedTransformerTokenizer,
                         SpacyTokenizer, WhitespaceTokenizer, LamboTokenizer)
from .dataset_readers import (ConllDatasetReader, DatasetReader,
                              TextClassificationJSONReader, UniversalDependenciesDatasetReader)
from .api import (Sentence, tokens2conllu, conllu2sentence, sentence2conllu)


class TokenCharactersIndexer:
    pass


from .token_embedder import *
from .empty_embedder import EmptyEmbedder
from .character_token_embedder import CharacterBasedWordEmbedder
from .pretrained_transformer_embedder import PretrainedTransformerEmbedder
from .pretrained_transformer_mismatched_embedder import PretrainedTransformerMismatchedEmbedder
from .transformers_words_embeddings import TransformersWordEmbedder
from .projected_words_embedder import ProjectedWordEmbedder



"""
Adapted from AllenNLP
"""
from typing import List

import torch
from overrides import overrides

from combo.config.registry import Registry
from combo.config.from_parameters import FromParameters, register_arguments


@Registry.register('time_distributed')
class TimeDistributed(torch.nn.Module, FromParameters):
    """
    Given an input shaped like `(batch_size, time_steps, [rest])` and a `Module` that takes
    inputs like `(batch_size, [rest])`, `TimeDistributed` reshapes the input to be
    `(batch_size * time_steps, [rest])`, applies the contained `Module`, then reshapes it back.

    Note that while the above gives shapes with `batch_size` first, this `Module` also works if
    `batch_size` is second - we always just combine the first two dimensions, then split them.

    It also reshapes keyword arguments unless they are not tensors or their name is specified in
    the optional `pass_through` iterable.
    """

    @register_arguments
    def __init__(self, module):
        super().__init__()
        self._module = module

    @overrides
    def forward(self, *inputs, pass_through: List[str] = None, **kwargs):

        pass_through = pass_through or []

        reshaped_inputs = [self._reshape_tensor(input_tensor) for input_tensor in inputs]

        # Need some input to then get the batch_size and time_steps.
        some_input = None
        if inputs:
            some_input = inputs[-1]

        reshaped_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, torch.Tensor) and key not in pass_through:
                if some_input is None:
                    some_input = value

                value = self._reshape_tensor(value)

            reshaped_kwargs[key] = value

        reshaped_outputs = self._module(*reshaped_inputs, **reshaped_kwargs)

        if some_input is None:
            raise RuntimeError("No input tensor to time-distribute")

        # Now get the output back into the right shape.
        # (batch_size, time_steps, **output_size)
        new_size = some_input.size()[:2] + reshaped_outputs.size()[1:]
        outputs = reshaped_outputs.contiguous().view(new_size)

        return outputs

    @staticmethod
    def _reshape_tensor(input_tensor):
        input_size = input_tensor.size()
        if len(input_size) <= 2:
            raise RuntimeError(f"No dimension to distribute: {input_size}")
        # Squash batch_size and time_steps into a single axis; result has shape
        # (batch_size * time_steps, **input_size).
        squashed_shape = [-1] + list(input_size[2:])
        return input_tensor.contiguous().view(*squashed_shape)



from typing import Optional, Dict, List, Union

import torch
import torch.nn as nn
from overrides import overrides

from combo import data
from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.models import dilated_cnn
from combo.nn import base
from combo.nn.activations import Activation
from combo.nn.utils import masked_cross_entropy
from combo.utils import ConfigurationError
from combo.models.time_distributed import TimeDistributed
from combo.predictors import Predictor


@Registry.register('combo_lemma_predictor_from_vocab')
class LemmatizerModel(Predictor):
    """Lemmatizer model."""

    @register_arguments
    def __init__(self,
                 vocabulary: data.Vocabulary,
                 char_vocab_namespace: str,
                 lemma_vocab_namespace: str,
                 embedding_dim: int,
                 input_projection_layer: base.Linear,
                 filters: List[int],
                 kernel_size: List[int],
                 stride: List[int],
                 padding: List[int],
                 dilation: List[int],
                 activations: List[Activation],
                 ):
        assert char_vocab_namespace in vocabulary.get_namespaces()
        assert lemma_vocab_namespace in vocabulary.get_namespaces()
        super().__init__()

        if len(filters) + 1 != len(kernel_size):
            raise ConfigurationError(
                f"len(filters) ({len(filters):d}) + 1 != kernel_size ({len(kernel_size):d})"
            )
        filters = filters + [vocabulary.get_vocab_size(lemma_vocab_namespace)]

        dilated_cnn_encoder = dilated_cnn.DilatedCnnEncoder(
            input_dim=embedding_dim + input_projection_layer.get_output_dim(),
            filters=filters,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            activations=activations,
        )
        self.char_embed = nn.Embedding(
            num_embeddings=vocabulary.get_vocab_size(char_vocab_namespace),
            embedding_dim=embedding_dim,
        )
        self.dilated_cnn_encoder = TimeDistributed(dilated_cnn_encoder)
        self.input_projection_layer = input_projection_layer

    @overrides
    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        encoder_emb, chars = x

        encoder_emb = self.input_projection_layer(encoder_emb)
        char_embeddings = self.char_embed(chars)

        BATCH_SIZE, _, MAX_WORD_LENGTH, CHAR_EMB = char_embeddings.size()
        encoder_emb = encoder_emb.unsqueeze(2).repeat(1, 1, MAX_WORD_LENGTH, 1)

        x = torch.cat((char_embeddings, encoder_emb), dim=-1).transpose(2, 3)
        x = self.dilated_cnn_encoder(x).transpose(2, 3)
        output = {
            "prediction": x.argmax(-1),
            "probability": x
        }

        if labels is not None:
            if mask is None:
                mask = encoder_emb.new_ones(encoder_emb.size()[:-2])
            if sample_weights is None:
                sample_weights = labels.new_ones(BATCH_SIZE)
            mask = mask.unsqueeze(2).repeat(1, 1, MAX_WORD_LENGTH).bool()
            output["loss"] = self._loss(x, labels, mask, sample_weights)

        return output

    @staticmethod
    def _loss(pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:
        BATCH_SIZE, SENTENCE_LENGTH, MAX_WORD_LENGTH, CHAR_CLASSES = pred.size()
        pred = pred.reshape(-1, CHAR_CLASSES)

        true = true.reshape(-1)
        mask = true.gt(0)
        loss = masked_cross_entropy(pred, true, mask)
        loss = loss.reshape(BATCH_SIZE, -1) * sample_weights.unsqueeze(-1)
        valid_positions = mask.sum()
        return loss.sum() / valid_positions



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/nn/initializers.py
"""
import itertools

import torch
from typing import List
from combo.utils import ConfigurationError


def block_orthogonal(tensor: torch.Tensor, split_sizes: List[int], gain: float = 1.0) -> None:
    """
    An initializer which allows initializing model parameters in "blocks". This is helpful
    in the case of recurrent models which use multiple gates applied to linear projections,
    which can be computed efficiently if they are concatenated together. However, they are
    separate parameters which should be initialized independently.
    # Parameters
    tensor : `torch.Tensor`, required.
        A tensor to initialize.
    split_sizes : `List[int]`, required.
        A list of length `tensor.ndim()` specifying the size of the
        blocks along that particular dimension. E.g. `[10, 20]` would
        result in the tensor being split into chunks of size 10 along the
        first dimension and 20 along the second.
    gain : `float`, optional (default = `1.0`)
        The gain (scaling) applied to the orthogonal initialization.
    """
    data = tensor.data
    sizes = list(tensor.size())
    if any(a % b != 0 for a, b in zip(sizes, split_sizes)):
        raise ConfigurationError(
            "tensor dimensions must be divisible by their respective "
            "split_sizes. Found size: {} and split_sizes: {}".format(sizes, split_sizes)
        )
    indexes = [list(range(0, max_size, split)) for max_size, split in zip(sizes, split_sizes)]
    # Iterate over all possible blocks within the tensor.
    for block_start_indices in itertools.product(*indexes):
        # A list of tuples containing the index to start at for this block
        # and the appropriate step size (i.e split_size[i] for dimension i).
        index_and_step_tuples = zip(block_start_indices, split_sizes)
        # This is a tuple of slices corresponding to:
        # tensor[index: index + step_size, ...]. This is
        # required because we could have an arbitrary number
        # of dimensions. The actual slices we need are the
        # start_index: start_index + step for each dimension in the tensor.
        block_slice = tuple(
            slice(start_index, start_index + step) for start_index, step in index_and_step_tuples
        )
        data[block_slice] = torch.nn.init.orthogonal_(tensor[block_slice].contiguous(), gain=gain)


import torch
import torch.nn as nn
from overrides import overrides

from combo.config import FromParameters
from combo.config.registry import Registry


class Activation(nn.Module, FromParameters):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


@Registry.register('linear')
class LinearActivation(Activation):
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


@Registry.register('relu')
class ReLUActivation(Activation):
    def __init__(self):
        super().__init__()
        self.__torch_activation = torch.nn.ReLU()
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__torch_activation.forward(x)


@Registry.register('leaky_relu')
class ReLUActivation(Activation):
    def __init__(self):
        super().__init__()
        self.__torch_activation = torch.nn.LeakyReLU()
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__torch_activation.forward(x)


@Registry.register('gelu')
class ReLUActivation(Activation):
    def __init__(self):
        super().__init__()
        self.__torch_activation = torch.nn.GELU()
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__torch_activation.forward(x)


@Registry.register('sigmoid')
class SigmoidActivation(Activation):
    def __init__(self):
        super().__init__()
        self.__torch_activation = torch.nn.Sigmoid()
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__torch_activation.forward(x)


@Registry.register('tanh')
class TanhActivation(Activation):
    def __init__(self):
        super().__init__()
        self.__torch_activation = torch.nn.Tanh()
    @overrides
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__torch_activation.forward(x)


@Registry.register('mish')
class MishActivation(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.tanh(torch.nn.functional.softplus(x))


@Registry.register('swish')
class SwishActivation(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.sigmoid(x)


@Registry.register('gelu_new')
class GeluNew(Activation):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT). Also
    see the Gaussian Error Linear Units paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return (
                0.5
                * x
                * (1.0 + torch.tanh(torch.math.sqrt(2.0 / torch.math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))
        )


@Registry.register('gelu_fast')
class GeluFast(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return 0.5 * x * (1.0 + torch.tanh(x * 0.7978845608 * (1.0 + 0.044715 * x * x)))



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/nn/regularizers/regularizer.py
"""

import torch

from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments


class Regularizer(FromParameters):
    """
    An abstract class representing a regularizer. It must implement
    call, returning a scalar tensor.
    """

    def __call__(self, parameter: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


@Registry.register('l1_regularizer')
class L1Regularizer(Regularizer):
    """
    Represents a penalty proportional to the sum of the absolute values of the parameters
    Registered as a `Regularizer` with name "l1".
    """
    @register_arguments
    def __init__(self, alpha: float = 0.01) -> None:
        self.alpha = alpha

    def __call__(self, parameter: torch.Tensor) -> torch.Tensor:
        return self.alpha * torch.sum(torch.abs(parameter))


@Registry.register('l2_regularizer')
class L2Regularizer(Regularizer):
    """
    Represents a penalty proportional to the sum of squared values of the parameters
    Registered as a `Regularizer` with name "l2".
    """

    @register_arguments
    def __init__(self, alpha: float = 0.01) -> None:
        self.alpha = alpha

    def __call__(self, parameter: torch.Tensor) -> torch.Tensor:
        return self.alpha * torch.sum(torch.pow(parameter, 2))



import unittest
import os

from combo.data.tokenizers import Token
from combo.data.token_indexers import TokenFeatsIndexer
from combo.data.vocabulary import Vocabulary


class TokenFeatsIndexerTest(unittest.TestCase):

    def setUp(self):
        self.vocabulary = Vocabulary.from_files(
            os.path.join(os.getcwd(), '../../fixtures/train_vocabulary'),
            oov_token='_',
            padding_token='__PAD__'
        )
        self.token_feats_indexer = TokenFeatsIndexer()

    def test_get_index_to_token(self):
        token = Token(idx=0, text='a', feats={'Mood': 'Ind'})
        counter = {'feats_labels': {'Mood=Ind': 0}}
        self.token_feats_indexer.count_vocab_items(token, counter)
        self.assertEqual(counter['feats_labels']['Mood=Ind'], 1)

    def test_tokens_to_indices(self):
        self.assertEqual(
            self.token_feats_indexer.tokens_to_indices(
                [Token(idx=0, text='a', feats={'Mood': 'Ind'})], self.vocabulary),
            {'tokens': [[11]+[0]*173]})

    def test_get_padding_lengths(self):
        self.assertEqual(
            self.token_feats_indexer.get_padding_lengths({'tokens': [[11]+[0]*173]}),
            {'tokens': 1}
        )

    def test_get_nondefault_padding_lengths(self):
        token_feats_indexer_w_padding = TokenFeatsIndexer(token_min_padding_length=4)
        self.assertEqual(
            token_feats_indexer_w_padding.get_padding_lengths({'tokens': [[11]+[0]*173]}),
            {'tokens': 4}
        )

    def test_tokens_to_indices_as_padded_tensor(self):
        token_feats_indexer_w_padding = TokenFeatsIndexer(token_min_padding_length=4)
        token = Token(idx=0, text='a', feats={'Mood': 'Ind'})
        indices_dict = token_feats_indexer_w_padding.tokens_to_indices([token], self.vocabulary)
        padded_dict = list(token_feats_indexer_w_padding.as_padded_tensor_dict(
            indices_dict,
            token_feats_indexer_w_padding.get_padding_lengths(indices_dict)
        )['tokens'].numpy()[0])
        self.assertListEqual(padded_dict, [11]+[0]*173)



"""
Author: Łukasz Pszenny
"""
from typing import Optional, List

from combo.common.tqdm import Tqdm
import numpy as np
import pandas as pd
from pathlib import Path

from combo.data import Sentence
from combo.utils import ComboLogger


def extract_combo_matrices(predictions: List[Sentence],
                           serialization_dir: Path,
                           input_data_path: Path,
                           logger: Optional[ComboLogger] = None,
                           logging_prefix: str = ''):
    OUTPUT_DIRECTORY_MATRICES = serialization_dir / "combo_output" / "dependency_tree_matrices"
    OUTPUT_RELATION_LABEL_DISTRIBUTION = serialization_dir / "combo_output" / "label_distribution_matrices"

    meta_ids = []
    meta_splits = []
    meta_file_names = []
    # For saving file names
    tmp_meta_file_names = []

    sentences = []
    sentence_ids = []

    # Create directory if it doesn't exist
    OUTPUT_DIRECTORY_MATRICES.mkdir(parents=True, exist_ok=True)
    OUTPUT_RELATION_LABEL_DISTRIBUTION.mkdir(parents=True, exist_ok=True)

    for sentence_id, predicted_sentence in Tqdm.tqdm(enumerate(predictions)):
        dependency_tree_matrix = predicted_sentence.relation_distribution
        np.savetxt(fname=OUTPUT_DIRECTORY_MATRICES / f'{input_data_path.stem}-{sentence_id}.csv',
                   X=dependency_tree_matrix,
                   delimiter=',')
        tmp_meta_file_names.append(f'{input_data_path.stem}-{sentence_id}.csv')

        # Save relation label matrix
        label_distribution_matrix = predicted_sentence.relation_label_distribution
        np.savetxt(fname=OUTPUT_RELATION_LABEL_DISTRIBUTION / f'{input_data_path.stem}-{sentence_id}.csv',
                   X=label_distribution_matrix,
                   delimiter=',')
        sentences.append(' '.join([t.text for t in predicted_sentence.tokens]))
        sentence_ids.append(sentence_id)
    logger.info(f"\nFinished processing : {input_data_path}", prefix=logging_prefix)

    meta_ids += sentence_ids
    meta_splits += [input_data_path.stem] * len(sentence_ids)
    meta_file_names += tmp_meta_file_names

    meta_data = pd.DataFrame({'split': meta_splits,
                              'sentence_id': meta_ids,
                              'full_sentence': sentences,
                              'file_names': meta_file_names})

    logger.info("Saving metadata", prefix=logging_prefix)
    meta_data.to_csv(serialization_dir / 'metadata.csv', index=False)
    logger.info("Finished processing", prefix=logging_prefix)



from copy import deepcopy
from typing import Dict, Generic, List, TypeVar, Any

import torch

from combo.data.vocabulary import Vocabulary


"""
Adapted from https://github.com/allenai/allennlp/blob/main/allennlp/data/fields/field.py
"""


DataArray = TypeVar(
    "DataArray", torch.Tensor, Dict[str, torch.Tensor], Dict[str, Dict[str, torch.Tensor]]
)


class Field(Generic[DataArray]):
    """
    A `Field` is some piece of a data instance that ends up as an tensor in a model (either as an
    input or an output).  Data instances are just collections of fields.
    Fields go through up to two steps of processing: (1) tokenized fields are converted into token
    ids, (2) fields containing token ids (or any other numeric data) are padded (if necessary) and
    converted into tensors.  The `Field` API has methods around both of these steps, though they
    may not be needed for some concrete `Field` classes - if your field doesn't have any strings
    that need indexing, you don't need to implement `count_vocab_items` or `index`.  These
    methods `pass` by default.
    Once a vocabulary is computed and all fields are indexed, we will determine padding lengths,
    then intelligently batch together instances and pad them into actual tensors.
    """

    __slots__ = []  # type: ignore

    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]):
        """
        If there are strings in this field that need to be converted into integers through a
        :class:`Vocabulary`, here is where we count them, to determine which tokens are in or out
        of the vocabulary.
        If your `Field` does not have any strings that need to be converted into indices, you do
        not need to implement this method.
        A note on this `counter`: because `Fields` can represent conceptually different things,
        we separate the vocabulary items by `namespaces`.  This way, we can use a single shared
        mechanism to handle all mappings from strings to integers in all fields, while keeping
        words in a `TextField` from sharing the same ids with labels in a `LabelField` (e.g.,
        "entailment" or "contradiction" are labels in an entailment task)
        Additionally, a single `Field` might want to use multiple namespaces - `TextFields` can
        be represented as a combination of word ids and character ids, and you don't want words and
        characters to share the same vocabulary - "a" as a word should get a different idx from "a"
        as a character, and the vocabulary sizes of words and characters are very different.
        Because of this, the first key in the `counter` object is a `_namespace`, like "tokens",
        "token_characters", "tags", or "labels", and the second key is the actual vocabulary item.
        """
        pass

    def human_readable_repr(self) -> Any:
        """
        This method should be implemented by subclasses to return a structured, yet human-readable
        representation of the field.
        !!! Note
            `human_readable_repr()` is not meant to be used as a method to serialize a `Field` since the return
            value does not necessarily contain all of the attributes of the `Field` instance. But the object
            returned should be JSON-serializable.
        """
        raise NotImplementedError

    def index(self, vocab: Vocabulary):
        """
        Given a :class:`Vocabulary`, converts all strings in this field into (typically) integers.
        This `modifies` the `Field` object, it does not return anything.
        If your `Field` does not have any strings that need to be converted into indices, you do
        not need to implement this method.
        """
        pass

    def get_padding_lengths(self) -> Dict[str, int]:
        """
        If there are things in this field that need padding, note them here.  In order to pad a
        batch of instance, we get all of the lengths from the batch, take the max, and pad
        everything to that length (or use a pre-specified maximum length).  The return value is a
        dictionary mapping keys to lengths, like `{'num_tokens': 13}`.
        This is always called after :func:`index`.
        """
        raise NotImplementedError

    def as_tensor(self, padding_lengths: Dict[str, int]) -> DataArray:
        """
        Given a set of specified padding lengths, actually pad the data in this field and return a
        torch Tensor (or a more complex data structure) of the correct shape.  We also take a
        couple of parameters that are important when constructing torch Tensors.
        # Parameters
        padding_lengths : `Dict[str, int]`
            This dictionary will have the same keys that were produced in
            :func:`get_padding_lengths`.  The values specify the lengths to use when padding each
            relevant dimension, aggregated across all instances in a batch.
        """
        raise NotImplementedError

    def empty_field(self) -> "Field":
        """
        So that `ListField` can pad the number of fields in a list (e.g., the number of answer
        option `TextFields`), we need a representation of an empty field of each type.  This
        returns that.  This will only ever be called when we're to the point of calling
        :func:`as_tensor`, so you don't need to worry about `get_padding_lengths`,
        `count_vocab_items`, etc., being called on this empty field.
        We make this an instance method instead of a static method so that if there is any state
        in the Field, we can copy it over (e.g., the token indexers in `TextField`).
        """
        raise NotImplementedError

    def batch_tensors(self, tensor_list: List[DataArray]) -> DataArray:  # type: ignore
        """
        Takes the output of `Field.as_tensor()` from a list of `Instances` and merges it into
        one batched tensor for this `Field`.  The default implementation here in the base class
        handles cases where `as_tensor` returns a single torch tensor per instance.  If your
        subclass returns something other than this, you need to override this method.
        This operation does not modify `self`, but in some cases we need the information
        contained in `self` in order to perform the batching, so this is an instance method, not
        a class method.
        """

        return torch.stack(tensor_list)

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            # With the way "slots" classes work, self.__slots__ only gives the slots defined
            # by the current class, but not any of its base classes. Therefore to truly
            # check for equality we have to check through all of the slots in all of the
            # base classes as well.
            for class_ in self.__class__.mro():
                for attr in getattr(class_, "__slots__", []):
                    if getattr(self, attr) != getattr(other, attr):
                        return False
            # It's possible that a subclass was not defined as a slots class, in which
            # case we'll need to check __dict__.
            if hasattr(self, "__dict__"):
                return self.__dict__ == other.__dict__
            return True
        return NotImplemented

    def __len__(self):
        raise NotImplementedError

    def duplicate(self):
        return deepcopy(self)


import unittest
import os

from combo.config import Registry
from combo.config.from_parameters import override_parameters
from combo.data import WhitespaceTokenizer, UniversalDependenciesDatasetReader, Vocabulary
from combo.data.token_indexers.token_characters_indexer import TokenCharactersIndexer


VOCABULARY_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '../fixtures/small_vocabulary'))


class ConfigurationTest(unittest.TestCase):

    def test_dataset_reader_from_registry(self):
        dr_type, constructor = Registry.resolve('conllu_dataset_reader')
        dataset_reader = dr_type.from_parameters()
        self.assertEqual(type(dataset_reader), UniversalDependenciesDatasetReader)

    def test_dataset_reader_from_registry_with_parameters(self):
        parameters = {'token_indexers': {'char': {'type': 'token_characters_indexer'}},
                      'use_sem': True}
        dr_type, constructor = Registry.resolve('conllu_dataset_reader')
        dataset_reader = dr_type.from_parameters(parameters, constructor)
        self.assertEqual(type(dataset_reader), UniversalDependenciesDatasetReader)
        self.assertEqual(type(dataset_reader.token_indexers['char']), TokenCharactersIndexer)
        self.assertEqual(dataset_reader.use_sem, True)

    def test_dataset_reader_from_registry_with_token_indexer_parameters(self):
        parameters = {'token_indexers': {'char': {'type': 'token_characters_indexer',
                                                  'parameters': {
                                                      'namespace': 'custom_namespace',
                                                      'tokenizer': {
                                                          'type': 'whitespace_tokenizer'
                                                      }
                                                  }}}}
        dr_type, constructor = Registry.resolve('conllu_dataset_reader')
        dataset_reader = dr_type.from_parameters(parameters)
        self.assertEqual(type(dataset_reader), UniversalDependenciesDatasetReader)
        self.assertEqual(dataset_reader.token_indexers['char']._namespace, 'custom_namespace')
        self.assertEqual(type(dataset_reader.token_indexers['char']._character_tokenizer), WhitespaceTokenizer)

    def test_vocabulary_from_files(self):
        parameters = {'type': 'from_files_vocabulary',
                      'parameters': {
                          'directory': VOCABULARY_DIR
                      }}
        vocab_type, constructor = Registry.resolve(parameters['type'])
        vocab = vocab_type.from_parameters(parameters['parameters'], constructor)
        self.assertEqual(type(vocab), Vocabulary)
        self.assertEqual(vocab.constructed_from, 'from_files')
        self.assertSetEqual(vocab.get_namespaces(), {'animals'})

    def test_serialize(self):
        vocab = Vocabulary({'counter': {'test': 0}}, max_vocab_size=10)
        self.assertDictEqual(vocab.serialize(),
                             {
                                 'type': 'base_vocabulary',
                                 'parameters': {
                                     'counter': {'counter': {'test': 0}},
                                     'max_vocab_size': 10
                                 }
                             })

    def test_serialize_non_base_constructor(self):
        vocab = Vocabulary.from_files(VOCABULARY_DIR)
        self.assertDictEqual(vocab.serialize(),
                             {
                                 'type': 'from_files_vocabulary',
                                 'parameters': {
                                     'directory': VOCABULARY_DIR
                                 }
                             })

    def test_serialize_and_load_non_base_constructor(self):
        vocab = Vocabulary.from_files(VOCABULARY_DIR)
        serialized_vocab = vocab.serialize()
        clz, constructor = Registry.resolve(serialized_vocab['type'])
        reconstructed_vocab = clz.from_parameters(serialized_vocab['parameters'], constructor)
        self.assertEqual(type(reconstructed_vocab), Vocabulary)
        self.assertEqual(reconstructed_vocab.constructed_from, 'from_files')
        self.assertSetEqual(reconstructed_vocab.get_namespaces(), {'animals'})


    def test_override_parameters(self):
        parameters = {
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}},
                'max_vocab_size': 10
            }
        }

        to_override = {'parameters': {'max_vocab_size': 15}}

        self.assertDictEqual({
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}},
                'max_vocab_size': 15
            }
        }, override_parameters(parameters, to_override))

    def test_override_nested_parameters(self):
        parameters = {
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}, 'another_property': 0},
                'another_counter': {'counter': {'test': 0}, 'another_property': 0}
            }
        }

        to_override = {'parameters': {'another_counter': {'counter': {'test': 1}}}}

        self.assertDictEqual({
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}, 'another_property': 0},
                'another_counter': {'counter': {'test': 1}, 'another_property': 0}
            }
        }, override_parameters(parameters, to_override))

    def test_override_parameters_no_change(self):
        parameters = {
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}},
                'max_vocab_size': 10
            }
        }

        to_override = {}

        self.assertDictEqual({
            'type': 'base_vocabulary',
            'parameters': {
                'counter': {'counter': {'test': 0}},
                'max_vocab_size': 10
            }
        }, override_parameters(parameters, to_override))



from typing import List, Optional

from lambo.segmenter.lambo import Lambo

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.tokenizers.token import Token
from combo.data.tokenizers.tokenizer import Tokenizer


@Registry.register('lambo_tokenizer')
class LamboTokenizer(Tokenizer):
    @register_arguments
    def __init__(
            self,
            language: str = "English",
            default_turns: bool = False,
            default_split_subwords: bool = False
    ):
        self._language = language
        self.__tokenizer = Lambo.get(language)
        self.__default_turns = default_turns
        self.__default_split_subwords = default_split_subwords

    def tokenize(self, text: str) -> List[Token]:
        """
        Simple tokenization - ignoring the sentence splits
        :param text:
        :return:
        """
        document = self.__tokenizer.segment(text)
        tokens = []

        for turn in document.turns:
            for sentence in turn.sentences:
                for token in sentence.tokens:
                    tokens.append(Token(token.text, subwords=token.subwords))

        return tokens

    def segment(self, text: str, turns: Optional[bool] = None, split_subwords: Optional[bool] = None) -> List[List[str]]:
        """
        Full segmentation - segment into sentences.
        :param text:
        :param turns: segment into sentences by splitting on sentences or on turns. Default: sentences.
        :return:
        """

        turns = turns or self.__default_turns
        split_subwords = split_subwords or self.__default_split_subwords

        document = self.__tokenizer.segment(text)
        sentences = []
        sentence_tokens = []

        for turn in document.turns:
            if turns:
                sentence_tokens = []
            for sentence in turn.sentences:
                if not turns:
                    sentence_tokens = []
                for token in sentence.tokens:
                    if len(token.subwords) > 0 and split_subwords:
                        sentence_tokens.extend([s for s in token.subwords])
                    else:
                        sentence_tokens.append(token.text)
                if not turns:
                    sentences.append(sentence_tokens)
            if turns:
                sentences.append(sentence_tokens)

        return sentences



import unittest
import os

from combo.data import ConllDatasetReader


class ConllDatasetReaderTest(unittest.TestCase):
    def test_read_all_tokens(self):
        reader = ConllDatasetReader(coding_scheme='IOB2')
        tokens = [token for token in reader.read(os.path.join(os.path.dirname(__file__), 'conll_test_file.txt'))]
        self.assertEqual(len(tokens), 6)

    def test_tokenize_correct_tokens(self):
        reader = ConllDatasetReader(coding_scheme='IOB2')
        token = next(iter(reader.read(os.path.join(os.path.dirname(__file__), 'conll_test_file.txt'))))
        self.assertListEqual([str(t) for t in token['tokens'].tokens],
                             ['SOCCER', '-', 'JAPAN', 'GET', 'LUCKY', 'WIN', ',',
                              'CHINA', 'IN', 'SURPRISE', 'DEFEAT', '.'])

    def test_tokenize_correct_tags(self):
        reader = ConllDatasetReader(coding_scheme='IOB2')
        token = next(iter(reader.read(os.path.join(os.path.dirname(__file__), 'conll_test_file.txt'))))
        self.assertListEqual(token['tags'].labels,
                             ['O', 'O', 'B-LOC', 'O', 'O', 'O', 'O',
                              'B-PER', 'O', 'O', 'O', 'O'])



"""
Adapted from COMBO.
Author: Mateusz Klimaszewski
"""

from typing import List, Optional, Union, Tuple, Dict

from combo import data
from combo.predictors import Predictor

import torch
import torch.nn.functional as F

from combo.nn.base import Linear


class GraphHeadPredictionModel(Predictor):
    """Head prediction model."""

    def __init__(self,
                 head_projection_layer: Linear,
                 dependency_projection_layer: Linear,
                 cycle_loss_n: int = 0,
                 graph_weighting: float = 0.2):
        super().__init__()
        self.head_projection_layer = head_projection_layer
        self.dependency_projection_layer = dependency_projection_layer
        self.cycle_loss_n = cycle_loss_n
        self.graph_weighting = graph_weighting

    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                mask: Optional[torch.BoolTensor] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        if mask is None:
            mask = x.new_ones(x.size()[-1])
        heads_labels = None
        if labels is not None and labels[0] is not None:
            heads_labels = labels

        head_arc_emb = self.head_projection_layer(x)
        dep_arc_emb = self.dependency_projection_layer(x)
        x = dep_arc_emb.bmm(head_arc_emb.transpose(2, 1))
        pred = x.sigmoid() > 0.5

        output = {
            "prediction": pred,
            "probability": x
        }

        if heads_labels is not None:
            if sample_weights is None:
                sample_weights = heads_labels.new_ones([mask.size(0)])
            output["loss"], output["cycle_loss"] = self._loss(x, heads_labels, mask, sample_weights)

        return output

    def _cycle_loss(self, pred: torch.Tensor):
        BATCH_SIZE, _, _ = pred.size()
        loss = pred.new_zeros(BATCH_SIZE)
        # Index from 1: as using non __ROOT__ tokens
        pred = pred.softmax(-1)[:, 1:, 1:]
        x = pred
        for i in range(self.cycle_loss_n):
            loss += self._batch_trace(x)

            # Don't multiple on last iteration
            if i < self.cycle_loss_n - 1:
                x = x.bmm(pred)

        return loss

    @staticmethod
    def _batch_trace(x: torch.Tensor) -> torch.Tensor:
        assert len(x.size()) == 3
        BATCH_SIZE, N, M = x.size()
        assert N == M
        identity = x.new_tensor(torch.eye(N))
        identity = identity.reshape((1, N, N))
        batch_identity = identity.repeat(BATCH_SIZE, 1, 1)
        return (x * batch_identity).sum((-1, -2))

    def _loss(self, pred: torch.Tensor, labels: torch.Tensor, mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        BATCH_SIZE, N, M = pred.size()
        assert N == M
        SENTENCE_LENGTH = N

        valid_positions = mask.sum()

        result = []
        true = labels
        # Ignore first pred dimension as it is ROOT token prediction
        for i in range(SENTENCE_LENGTH - 1):
            pred_i = pred[:, i + 1, 1:].reshape(-1)
            true_i = true[:, i + 1, 1:].reshape(-1)
            mask_i = mask[:, i]
            bce_loss = F.binary_cross_entropy_with_logits(pred_i, true_i, reduction="none").mean(-1) * mask_i
            result.append(bce_loss)
        cycle_loss = self._cycle_loss(pred)
        loss = torch.stack(result).transpose(1, 0) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions + cycle_loss.mean(), cycle_loss.mean()


class GraphDependencyRelationModel(Predictor):
    """Dependency relation parsing model."""

    def __init__(self,
                 head_predictor: GraphHeadPredictionModel,
                 head_projection_layer: Linear,
                 dependency_projection_layer: Linear,
                 relation_prediction_layer: Linear):
        super().__init__()
        self.head_predictor = head_predictor
        self.head_projection_layer = head_projection_layer
        self.dependency_projection_layer = dependency_projection_layer
        self.relation_prediction_layer = relation_prediction_layer

    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        relations_labels, head_labels, enhanced_heads_labels, enhanced_deprels_labels = None, None, None, None
        if labels is not None and labels[0] is not None:
            relations_labels, head_labels, enhanced_heads_labels = labels

        head_output = self.head_predictor(x, enhanced_heads_labels, mask, sample_weights)
        head_pred = head_output["probability"]
        BATCH_SIZE, LENGTH, _ = head_pred.size()

        head_rel_emb = self.head_projection_layer(x)

        dep_rel_emb = self.dependency_projection_layer(x)

        # All possible edges combinations for each batch
        # Repeat interleave to have [emb1, emb1 ... (length times) ... emb1, emb2 ... ]
        head_rel_pred = head_rel_emb.repeat_interleave(LENGTH, -2)
        # Regular repeat to have all combinations [deprel1, deprel2, ... deprelL, deprel1 ...]
        dep_rel_pred = dep_rel_emb.repeat(1, LENGTH, 1)

        # All possible edges combinations for each batch
        dep_rel_pred = torch.cat((head_rel_pred, dep_rel_pred), dim=-1)

        relation_prediction = self.relation_prediction_layer(dep_rel_pred).reshape(BATCH_SIZE, LENGTH, LENGTH, -1)
        output = head_output

        output["prediction"] = (relation_prediction.argmax(-1), head_output["prediction"])
        output["rel_probability"] = relation_prediction

        if labels is not None and labels[0] is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            loss = self._loss(relation_prediction, relations_labels, enhanced_heads_labels, mask, sample_weights)
            output["loss"] = (loss, head_output["loss"])

        return output

    @staticmethod
    def _loss(pred: torch.Tensor,
              true: torch.Tensor,
              heads_true: torch.Tensor,
              mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:
        correct_heads_mask = heads_true.long() == 1
        true = true[correct_heads_mask]
        pred = pred[correct_heads_mask]
        loss = F.cross_entropy(pred, true.long())
        return loss.sum() / pred.size(0)

    @classmethod
    def from_vocab(cls,
                   vocab: data.Vocabulary,
                   vocab_namespace: str,
                   head_predictor: GraphHeadPredictionModel,
                   head_projection_layer: Linear,
                   dependency_projection_layer: Linear
                   ):
        """Creates parser combining model configuration and vocabulary data."""
        assert vocab_namespace in vocab.get_namespaces()
        relation_prediction_layer = Linear(
            in_features=head_projection_layer.get_output_dim() + dependency_projection_layer.get_output_dim(),
            out_features=vocab.get_vocab_size(vocab_namespace)
        )
        return cls(
            head_predictor=head_predictor,
            head_projection_layer=head_projection_layer,
            dependency_projection_layer=dependency_projection_layer,
            relation_prediction_layer=relation_prediction_layer
        )



"""
Partially adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/data/dataset_readers/dataset_utils/span_utils.py
"""

from typing import Callable, List, Set, Tuple, TypeVar, Optional

from combo.data.tokenizers.tokenizer import Token
from combo.utils import ConfigurationError, InvalidTagSequence

TypedSpan = Tuple[int, Tuple[int, int]]
TypedStringSpan = Tuple[str, Tuple[int, int]]

T = TypeVar("T", str, Token)


def bio_tags_to_spans(
        tag_sequence: List[str], classes_to_ignore: List[str] = None
) -> List[TypedStringSpan]:
    """
    Given a sequence corresponding to BIO tags, extracts spans.
    Spans are inclusive and can be of zero length, representing a single word span.
    Ill-formed spans are also included (i.e those which do not start with a "B-LABEL"),
    as otherwise it is possible to get a perfect precision score whilst still predicting
    ill-formed spans in addition to the correct spans. This function works properly when
    the spans are unlabeled (i.e., your labels are simply "B", "I", and "O").
    # Parameters
    tag_sequence : `List[str]`, required.
        The integer class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
        Note that the label `does not` contain any BIO tag prefixes.
    """
    classes_to_ignore = classes_to_ignore or []
    spans: Set[Tuple[str, Tuple[int, int]]] = set()
    span_start = 0
    span_end = 0
    active_conll_tag = None
    for index, string_tag in enumerate(tag_sequence):
        # Actual BIO tag.
        bio_tag = string_tag[0]
        if bio_tag not in ["B", "I", "O"]:
            raise InvalidTagSequence(tag_sequence)
        conll_tag = string_tag[2:]
        if bio_tag == "O" or conll_tag in classes_to_ignore:
            # The span has ended.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = None
            # We don't care about tags we are
            # told to ignore, so we do nothing.
            continue
        elif bio_tag == "B":
            # We are entering a new span; reset indices
            # and active tag to new span.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = conll_tag
            span_start = index
            span_end = index
        elif bio_tag == "I" and conll_tag == active_conll_tag:
            # We're inside a span.
            span_end += 1
        else:
            # This is the case the bio label is an "I", but either:
            # 1) the span hasn't started - i.e. an ill formed span.
            # 2) The span is an I tag for a different conll annotation.
            # We'll process the previous span if it exists, but also
            # include this span. This is important, because otherwise,
            # a model may get a perfect F1 score whilst still including
            # false positive ill-formed spans.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = conll_tag
            span_start = index
            span_end = index
    # Last token might have been a part of a valid span.
    if active_conll_tag is not None:
        spans.add((active_conll_tag, (span_start, span_end)))
    return list(spans)


def _iob_tags_to_spans(
        start_of_chunk_fun: Callable[[Optional[str], Optional[str], str, str], bool],
        tag_sequence: List[str], classes_to_ignore: List[str] = None,
) -> List[TypedStringSpan]:
    """
    Given a sequence corresponding to IOB1 tags, extracts spans.
    Spans are inclusive and can be of zero length, representing a single word span.
    Ill-formed spans are also included (i.e., those where "B-LABEL" is not preceded
    by "I-LABEL" or "B-LABEL").
    # Parameters
    tag_sequence : `List[str]`, required.
        The integer class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
        Note that the label `does not` contain any BIO tag prefixes.
    """
    classes_to_ignore = classes_to_ignore or []
    spans: Set[Tuple[str, Tuple[int, int]]] = set()
    span_start = 0
    span_end = 0
    active_conll_tag = None
    prev_bio_tag = None
    prev_conll_tag = None
    for index, string_tag in enumerate(tag_sequence):
        curr_bio_tag = string_tag[0]
        curr_conll_tag = string_tag[2:]

        if curr_bio_tag not in ["B", "I", "O"]:
            raise InvalidTagSequence(tag_sequence)
        if curr_bio_tag == "O" or curr_conll_tag in classes_to_ignore:
            # The span has ended.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = None
        elif start_of_chunk_fun(prev_bio_tag, prev_conll_tag, curr_bio_tag, curr_conll_tag):
            # We are entering a new span; reset indices
            # and active tag to new span.
            if active_conll_tag is not None:
                spans.add((active_conll_tag, (span_start, span_end)))
            active_conll_tag = curr_conll_tag
            span_start = index
            span_end = index
        else:
            # bio_tag == "I" and curr_conll_tag == active_conll_tag
            # We're continuing a span.
            span_end += 1

        prev_bio_tag = string_tag[0]
        prev_conll_tag = string_tag[2:]
    # Last token might have been a part of a valid span.
    if active_conll_tag is not None:
        spans.add((active_conll_tag, (span_start, span_end)))
    return list(spans)


def _iob1_start_of_chunk(
        prev_bio_tag: Optional[str],
        prev_conll_tag: Optional[str],
        curr_bio_tag: str,
        curr_conll_tag: str,
) -> bool:
    if curr_bio_tag == "B":
        return True
    if curr_bio_tag == "I" and prev_bio_tag == "O":
        return True
    if curr_bio_tag != "O" and prev_conll_tag != curr_conll_tag:
        return True
    return False


def _iob2_start_of_chunk(
        prev_bio_tag: Optional[str],
        prev_conll_tag: Optional[str],
        curr_bio_tag: str,
        curr_conll_tag: str,
) -> bool:
    if curr_bio_tag == "B":
        return True
    if curr_bio_tag != "O" and prev_conll_tag != curr_conll_tag:
        return True
    return False


def iob1_tags_to_spans(
        tag_sequence: List[str], classes_to_ignore: List[str] = None,
) -> List[TypedStringSpan]:
    return _iob_tags_to_spans(_iob1_start_of_chunk,
                              tag_sequence, classes_to_ignore)


def iob2_tags_to_spans(
        tag_sequence: List[str], classes_to_ignore: List[str] = None,
) -> List[TypedStringSpan]:
    return _iob_tags_to_spans(_iob2_start_of_chunk,
                              tag_sequence, classes_to_ignore)


def bioul_tags_to_spans(
        tag_sequence: List[str], classes_to_ignore: List[str] = None
) -> List[TypedStringSpan]:
    """
    Given a sequence corresponding to BIOUL tags, extracts spans.
    Spans are inclusive and can be of zero length, representing a single word span.
    Ill-formed spans are not allowed and will raise `InvalidTagSequence`.
    This function works properly when the spans are unlabeled (i.e., your labels are
    simply "B", "I", "O", "U", and "L").
    # Parameters
    tag_sequence : `List[str]`, required.
        The tag sequence encoded in BIOUL, e.g. ["B-PER", "L-PER", "O"].
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
    """
    spans = []
    classes_to_ignore = classes_to_ignore or []
    index = 0
    while index < len(tag_sequence):
        label = tag_sequence[index]
        if label[0] == "U":
            spans.append((label.partition("-")[2], (index, index)))
        elif label[0] == "B":
            start = index
            while label[0] != "L":
                index += 1
                if index >= len(tag_sequence):
                    raise InvalidTagSequence(tag_sequence)
                label = tag_sequence[index]
                if not (label[0] == "I" or label[0] == "L"):
                    raise InvalidTagSequence(tag_sequence)
            spans.append((label.partition("-")[2], (start, index)))
        else:
            if label != "O":
                raise InvalidTagSequence(tag_sequence)
        index += 1
    return [span for span in spans if span[0] not in classes_to_ignore]


def to_bioul(tag_sequence: List[str], encoding: str = "IOB1") -> List[str]:
    """
    Given a tag sequence encoded with IOB1 labels, recode to BIOUL.
    In the IOB1 scheme, I is a token inside a span, O is a token outside
    a span and B is the beginning of span immediately following another
    span of the same type.
    In the BIO scheme, I is a token inside a span, O is a token outside
    a span and B is the beginning of a span.
    # Parameters
    tag_sequence : `List[str]`, required.
        The tag sequence encoded in IOB1, e.g. ["I-PER", "I-PER", "O"].
    encoding : `str`, optional, (default = `"IOB1"`).
        The encoding type to convert from. Must be either "IOB1", "IOB2", or "BIO".
    # Returns
    bioul_sequence : `List[str]`
        The tag sequence encoded in IOB1, e.g. ["B-PER", "L-PER", "O"].
    """
    if encoding not in {"IOB1", "IOB2", "BIO"}:
        raise ConfigurationError(f"Invalid encoding {encoding} passed to 'to_bioul'.")

    def replace_label(full_label, new_label):
        # example: full_label = 'I-PER', new_label = 'U', returns 'U-PER'
        parts = list(full_label.partition("-"))
        parts[0] = new_label
        return "".join(parts)

    def pop_replace_append(in_stack, out_stack, new_label):
        # pop the last element from in_stack, replace the label, append
        # to out_stack
        tag = in_stack.pop()
        new_tag = replace_label(tag, new_label)
        out_stack.append(new_tag)

    def process_stack(stack, out_stack):
        # process a stack of labels, add them to out_stack
        if len(stack) == 1:
            # just a U token
            pop_replace_append(stack, out_stack, "U")
        else:
            # need to code as BIL
            recoded_stack = []
            pop_replace_append(stack, recoded_stack, "L")
            while len(stack) >= 2:
                pop_replace_append(stack, recoded_stack, "I")
            pop_replace_append(stack, recoded_stack, "B")
            recoded_stack.reverse()
            out_stack.extend(recoded_stack)

    # Process the tag_sequence one tag at a time, adding spans to a stack,
    # then recode them.
    bioul_sequence = []
    stack: List[str] = []

    for label in tag_sequence:
        # need to make a dict like
        # token = {'token': 'Matt', "labels": {'conll2003': "B-PER"}
        #                   'gold': 'I-PER'}
        # where 'gold' is the raw value from the CoNLL data set

        if label == "O" and len(stack) == 0:
            bioul_sequence.append(label)
        elif label == "O" and len(stack) > 0:
            # need to process the entries on the stack plus this one
            process_stack(stack, bioul_sequence)
            bioul_sequence.append(label)
        elif label[0] == "I":
            # IOB1:
            # check if the previous type is the same as this one
            # if it is then append to stack
            # otherwise this start a new entity if the type
            # is different
            if len(stack) == 0:
                # Beginning of the sequence
                if encoding in {"IOB2", "BIO"}:
                    raise InvalidTagSequence(tag_sequence)
                stack.append(label)
            else:
                # check if the previous type is the same as this one
                this_type = label.partition("-")[2]
                prev_type = stack[-1].partition("-")[2]
                if this_type == prev_type:
                    stack.append(label)
                else:
                    if encoding in {"IOB2", "BIO"}:
                        raise InvalidTagSequence(tag_sequence)
                    # a new entity
                    process_stack(stack, bioul_sequence)
                    stack.append(label)
        elif label[0] == "B":
            if len(stack) > 0:
                process_stack(stack, bioul_sequence)
            stack.append(label)
        else:
            raise InvalidTagSequence(tag_sequence)

    # process the stack
    if len(stack) > 0:
        process_stack(stack, bioul_sequence)

    return bioul_sequence


def bmes_tags_to_spans(
        tag_sequence: List[str], classes_to_ignore: List[str] = None
) -> List[TypedStringSpan]:
    """
    Given a sequence corresponding to BMES tags, extracts spans.
    Spans are inclusive and can be of zero length, representing a single word span.
    Ill-formed spans are also included (i.e those which do not start with a "B-LABEL"),
    as otherwise it is possible to get a perfect precision score whilst still predicting
    ill-formed spans in addition to the correct spans.
    This function works properly when the spans are unlabeled (i.e., your labels are
    simply "B", "M", "E" and "S").
    # Parameters
    tag_sequence : `List[str]`, required.
        The integer class labels for a sequence.
    classes_to_ignore : `List[str]`, optional (default = `None`).
        A list of string class labels `excluding` the bio tag
        which should be ignored when extracting spans.
    # Returns
    spans : `List[TypedStringSpan]`
        The typed, extracted spans from the sequence, in the format (label, (span_start, span_end)).
        Note that the label `does not` contain any BIO tag prefixes.
    """

    def extract_bmes_tag_label(text):
        bmes_tag = text[0]
        label = text[2:]
        return bmes_tag, label

    spans: List[Tuple[str, List[int]]] = []
    prev_bmes_tag: Optional[str] = None
    for index, tag in enumerate(tag_sequence):
        bmes_tag, label = extract_bmes_tag_label(tag)
        if bmes_tag in ("B", "S"):
            # Regardless of tag, we start a new span when reaching B & S.
            spans.append((label, [index, index]))
        elif bmes_tag in ("M", "E") and prev_bmes_tag in ("B", "M") and spans[-1][0] == label:
            # Only expand the span if
            # 1. Valid transition: B/M -> M/E.
            # 2. Matched label.
            spans[-1][1][1] = index
        else:
            # Best effort split for invalid span.
            spans.append((label, [index, index]))
        # update previous BMES tag.
        prev_bmes_tag = bmes_tag

    classes_to_ignore = classes_to_ignore or []
    return [
        # to tuple.
        (span[0], (span[1][0], span[1][1]))
        for span in spans
        if span[0] not in classes_to_ignore
    ]



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/data/token_indexers/single_id_token_indexer.py
"""

import itertools
from typing import Dict, List, Optional

from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.token_indexers import TokenIndexer, IndexedTokenList
from combo.data.tokenizers import Token
from combo.data.vocabulary import Vocabulary

_DEFAULT_VALUE = "THIS IS A REALLY UNLIKELY VALUE THAT HAS TO BE A STRING"


@Registry.register('single_id_token_indexer')
class SingleIdTokenIndexer(TokenIndexer):
    """
    This :class:`TokenIndexer` represents tokens as single integers.
    Registered as a `TokenIndexer` with name "single_id".
    # Parameters
    _namespace : `Optional[str]`, optional (default=`"tokens"`)
        We will use this _namespace in the :class:`Vocabulary` to map strings to indices.  If you
        explicitly pass in `None` here, we will skip indexing and vocabulary lookups.  This means
        that the `feature_name` you use must correspond to an integer value (like `text_id`, for
        instance, which gets set by some tokenizers, such as when using byte encoding).
    _lowercase_tokens : `bool`, optional (default=`False`)
        If `True`, we will call `token.lower()` before getting an index for the token from the
        vocabulary.
    start_tokens : `List[str]`, optional (default=`None`)
        These are prepended to the tokens provided to `tokens_to_indices`.
    end_tokens : `List[str]`, optional (default=`None`)
        These are appended to the tokens provided to `tokens_to_indices`.
    feature_name : `str`, optional (default=`"text"`)
        We will use the :class:`Token` attribute with this name as input.  This is potentially
        useful, e.g., for using NER tags instead of (or in addition to) surface forms as your inputs
        (passing `ent_type_` here would do that).  If you use a non-default value here, you almost
        certainly want to also change the `_namespace` parameter, and you might want to give a
        `default_value`.
    default_value : `str`, optional
        When you want to use a non-default `feature_name`, you sometimes want to have a default
        value to go with it, e.g., in case you don't have an NER tag for a particular token, for
        some reason.  This value will get used if we don't find a value in `feature_name`.  If this
        is not given, we will crash if a token doesn't have a value for the given `feature_name`, so
        that you don't get weird, silent errors by default.
    token_min_padding_length : `int`, optional (default=`0`)
        See :class:`TokenIndexer`.
    """

    @register_arguments
    def __init__(
            self,
            namespace: Optional[str] = "tokens",
            lowercase_tokens: bool = False,
            start_tokens: List[str] = None,
            end_tokens: List[str] = None,
            feature_name: str = "text",
            default_value: str = _DEFAULT_VALUE,
            token_min_padding_length: int = 0,
    ) -> None:
        super().__init__(token_min_padding_length)
        self._namespace = namespace
        self._lowercase_tokens = lowercase_tokens

        self._start_tokens = [Token(st) for st in (start_tokens or [])]
        self._end_tokens = [Token(et) for et in (end_tokens or [])]
        self._feature_name = feature_name
        self._default_value = default_value

    @overrides
    def count_vocab_items(self, token: Token, counter: Dict[str, Dict[str, int]]):
        if self._namespace is not None:
            text = self._get_feature_value(token)
            if self._lowercase_tokens:
                text = text.lower()
            counter[self._namespace][text] += 1

    @overrides
    def tokens_to_indices(
            self, tokens: List[Token], vocabulary: Vocabulary
    ) -> IndexedTokenList:
        indices: List[int] = []

        for token in itertools.chain(self._start_tokens, tokens, self._end_tokens):
            text = self._get_feature_value(token)
            if self._namespace is None:
                # We could have a check here that `text` is an int; not sure it's worth it.
                indices.append(text)  # type: ignore
            else:
                if self._lowercase_tokens:
                    text = text.lower()
                indices.append(vocabulary.get_token_index(text, self._namespace))

        return {"tokens": indices}

    @overrides
    def get_empty_token_list(self) -> IndexedTokenList:
        return {"tokens": []}

    def _get_feature_value(self, token: Token) -> str:
        text = getattr(token, self._feature_name)
        if text is None:
            if self._default_value is not _DEFAULT_VALUE:
                text = self._default_value
            else:
                raise ValueError(
                    f"{token} did not have attribute {self._feature_name}. If you "
                    "want to ignore this kind of error, give a default value in the "
                    "constructor of this indexer."
                )
        return text




from .tokenizer import Tokenizer, Token
from .character_tokenizer import CharacterTokenizer
from .pretrained_transformer_tokenizer import PretrainedTransformerTokenizer
from .spacy_tokenizer import SpacyTokenizer
from .sentence_splitter import SentenceSplitter, SpacySentenceSplitter
from .whitespace_tokenizer import WhitespaceTokenizer
from .lambo_tokenizer import LamboTokenizer



"""
Adapted from AllenNLP
Author: Mateusz Klimaszewski
"""
import copy
import pathlib
from abc import ABC
from typing import List, Any, Dict, Iterable, Tuple, Optional

import conllu
import torch
from conllu import parser
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary, fields, Instance, Token
from combo.data.dataset_readers.dataset_reader import DatasetReader, DatasetReaderInput
from combo.data.fields import Field
from combo.data.fields.adjacency_field import AdjacencyField
from combo.data.fields.metadata_field import MetadataField
from combo.data.fields.sequence_label_field import SequenceLabelField
from combo.data.fields.text_field import TextField
from combo.data.token_indexers import TokenIndexer
from combo.data.vocabulary import get_slices_if_not_provided
from combo.utils import checks, pad_sequence_to_length
from combo.data import Tokenizer


def parser_field_to_dataset_reader_field(field: str):
    if field == 'upos':
        return 'upostag'
    elif field == 'xpos':
        return 'xpostag'
    else:
        return field


@Registry.register('conllu_dataset_reader')
class UniversalDependenciesDatasetReader(DatasetReader, ABC):
    @register_arguments
    def __init__(
            self,
            tokenizer: Optional[Tokenizer] = None,
            token_indexers: Dict[str, TokenIndexer] = None,
            lemma_indexers: Dict[str, TokenIndexer] = None,
            features: List[str] = None,
            targets: List[str] = None,
            use_sem: bool = False,
            **kwargs,
    ) -> None:
        super().__init__(tokenizer=tokenizer, token_indexers=token_indexers)
        if features is None:
            features = ["token", "char"]
        if targets is None:
            targets = ["head", "deprel", "upostag", "xpostag", "lemma", "feats"]

        if "token" not in features and "char" not in features:
            raise checks.ConfigurationError("There must be at least one ('char' or 'token') text-based feature!")

        if "deps" in targets and not ("head" in targets and "deprel" in targets):
            raise checks.ConfigurationError("Add 'head' and 'deprel' to targets when using 'deps'!")

        intersection = set(features).intersection(set(targets))
        if len(intersection) != 0:
            raise checks.ConfigurationError(
                "Features and targets cannot share elements! "
                "Remove {} from either features or targets.".format(intersection)
            )
        self.use_sem = use_sem

        # *.conllu readers configuration
        fields = list(parser.DEFAULT_FIELDS)
        fields[1] = "token"  # use 'token' instead of 'form'
        field_parsers = parser.DEFAULT_FIELD_PARSERS
        # Do not make it nullable
        field_parsers.pop("xpostag", None)
        # Ignore parsing misc
        field_parsers.pop("misc", None)
        if self.use_sem:
            fields = list(fields)
            fields.append("semrel")
            field_parsers["semrel"] = lambda line, i: line[i]

        self.field_parsers = field_parsers
        self.fields = tuple([parser_field_to_dataset_reader_field(f) for f in fields])

        self.__lemma_indexers = lemma_indexers
        self.__targets = targets
        self.__features = features
        self.generate_labels = True
        # Filter out not required token indexers to avoid
        # Mismatched token keys ConfigurationError
        for indexer_name in list(self.token_indexers.keys()):
            if indexer_name not in self.features:
                del self.token_indexers[indexer_name]

    @property
    def lemma_indexers(self) -> Dict[str, TokenIndexer]:
        return self.__lemma_indexers

    @lemma_indexers.setter
    def lemma_indexers(self, value: Dict[str, TokenIndexer]):
        self.__lemma_indexers = value

    @property
    def targets(self) -> List[str]:
        return self.__targets

    @targets.setter
    def targets(self, value: List[str]):
        self.__targets = value

    @property
    def features(self) -> List[str]:
        return self.__features

    @features.setter
    def features(self, value: List[str]):
        self.__features = value

    @overrides
    def _read(self, file_path: DatasetReaderInput) -> Iterable[Instance]:
        file_path = [file_path] if len(file_path.split(",")) == 0 else file_path.split(",")

        for conllu_file in file_path:
            file = pathlib.Path(conllu_file)
            assert conllu_file and file.exists(), f"File with path '{conllu_file}' does not exist!"
            with file.open("r", encoding="utf-8") as f:
                for annotation in conllu.parse_incr(f, fields=self.fields, field_parsers=self.field_parsers):
                    yield self.text_to_instance(annotation)


    def text_to_instance(self, tree: conllu.models.TokenList) -> Instance:
        fields_: Dict[str, Field] = {}
        tokens = [Token.from_conllu_token(t) for t in tree if isinstance(t["id"], int)]

        # features
        text_field = TextField(tokens, self.token_indexers)
        fields_["sentence"] = text_field

        # targets
        if self.generate_labels:
            for target_name in self.targets:
                if target_name != "sent":
                    target_values = [t[target_name] for t in tokens]
                    if target_name == "lemma":
                        target_values = [Token(v) for v in target_values]
                        fields_[target_name] = TextField(target_values, self.lemma_indexers)
                    elif target_name == "feats":
                        target_values = self._feat_values(tokens)
                        fields_[target_name] = fields.SequenceMultiLabelField(target_values,
                                                                              self._feats_indexer,
                                                                              self._feats_as_tensor_wrapper,
                                                                              text_field,
                                                                              label_namespace="feats_labels")
                    elif target_name == "head":
                        target_values = [0 if v == "_" else int(v) for v in target_values]
                        fields_[target_name] = SequenceLabelField(target_values, text_field,
                                                                  label_namespace=target_name + "_labels")
                    elif target_name == "deps":
                        # Graphs require adding ROOT (AdjacencyField uses sequence length from TextField).
                        text_field_deps = TextField([Token("ROOT")] + copy.deepcopy(tokens), self._token_indexers)
                        enhanced_heads: List[Tuple[int, int]] = []
                        enhanced_deprels: List[str] = []
                        for idx, t in enumerate(tokens):
                            t_deps = t["deps"]
                            if t_deps and t_deps != "_":
                                for rel, head in t_deps:
                                    # EmoryNLP skips the first edge, if there are two edges between the same
                                    # nodes. Thanks to that one is in a tree and another in a graph.
                                    # This snippet follows that approach.
                                    if enhanced_heads and enhanced_heads[-1] == (idx, head):
                                        enhanced_heads.pop()
                                        enhanced_deprels.pop()
                                    enhanced_heads.append((idx, head))
                                    enhanced_deprels.append(rel)
                        fields_["enhanced_heads"] = AdjacencyField(
                            indices=enhanced_heads,
                            sequence_field=text_field_deps,
                            label_namespace="enhanced_heads_labels",
                            padding_value=0,
                        )
                        fields_["enhanced_deprels"] = AdjacencyField(
                            indices=enhanced_heads,
                            sequence_field=text_field_deps,
                            labels=enhanced_deprels,
                            # Label _namespace matches regular tree parsing.
                            label_namespace="enhanced_deprel_labels",
                            padding_value=0,
                        )
                    else:
                        # TODO: co robic gdy nie ma xpostag?
                        if all([t is None for t in target_values]):
                            continue
                        fields_[target_name] = SequenceLabelField(target_values, text_field,
                                                                  label_namespace=target_name + "_labels")

        # Restore feats fields to string representation
        # parser.serialize_field doesn't handle key without value
        for token in tokens:
            if token.feats:
                feats = token["feats"]
                if feats:
                    feats_values = []
                    for k, v in feats.items():
                        feats_values.append('='.join((k, v)) if v else k)
                    field = "|".join(feats_values)
                else:
                    field = "_"
                token.feats = field

        # metadata
        fields_["metadata"] = MetadataField({"input": tokens,
                                             "field_names": self.fields,
                                             "tokens": tokens})

        return Instance(fields_)

    @staticmethod
    def _feat_values(tree: List[Dict[str, Any]]):
        features = []
        for token in tree:
            token_features = []
            if token["feats"] is not None:
                for feat, value in token["feats"].items():
                    if feat in ["_", "__ROOT__"]:
                        pass
                    else:
                        # Handle case where feature is binary (doesn't have associated value)
                        if value:
                            token_features.append(feat + "=" + value)
                        else:
                            token_features.append(feat)
            features.append(token_features)
        return features

    @staticmethod
    def _feats_as_tensor_wrapper(field: fields.SequenceMultiLabelField):
        def as_tensor(padding_lengths):
            desired_num_tokens = padding_lengths["num_tokens"]
            assert len(field._indexed_multi_labels) > 0
            classes_count = len(field._indexed_multi_labels[0])
            default_value = [0.0] * classes_count
            padded_tags = pad_sequence_to_length(field._indexed_multi_labels, desired_num_tokens,
                                                 lambda: default_value)
            tensor = torch.tensor(padded_tags, dtype=torch.long)
            return tensor

        return as_tensor

    @staticmethod
    def _feats_indexer(vocab: Vocabulary):
        label_namespace = "feats_labels"
        vocab_size = vocab.get_vocab_size(label_namespace)
        slices = get_slices_if_not_provided(vocab)

        def _m_from_n_ones_encoding(multi_label: List[str], sentence_length: int) -> List[int]:
            one_hot_encoding = [0] * vocab_size
            for cat, cat_indices in slices.items():
                if cat not in ["__PAD__", "_"]:
                    label_from_cat = [label for label in multi_label if cat == label.split("=")[0]]
                    if label_from_cat:
                        label_from_cat = label_from_cat[0]
                        index = vocab.get_token_index(label_from_cat, label_namespace)
                    else:
                        # Get Cat=None index
                        index = vocab.get_token_index(cat + "=None", label_namespace)
                    one_hot_encoding[index] = 1
            return one_hot_encoding

        return _m_from_n_ones_encoding



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/modules/token_embedders/embedding.py
"""
import io
import itertools
import logging
import re
import tarfile
import zipfile
from typing import Any, Iterator, Tuple, NamedTuple, Optional, BinaryIO, Sequence

from combo.utils.typing import cast
from combo.utils.file_utils import cached_path, get_file_extension

logger = logging.getLogger(__name__)


class EmbeddingsFileURI(NamedTuple):
    main_file_uri: str
    path_inside_archive: Optional[str] = None


def format_embeddings_file_uri(
    main_file_path_or_url: str, path_inside_archive: Optional[str] = None
) -> str:
    if path_inside_archive:
        return "({})#{}".format(main_file_path_or_url, path_inside_archive)
    return main_file_path_or_url


def parse_embeddings_file_uri(uri: str) -> "EmbeddingsFileURI":
    match = re.fullmatch(r"\((.*)\)#(.*)", uri)
    if match:
        fields = cast(Tuple[str, str], match.groups())
        return EmbeddingsFileURI(*fields)
    else:
        return EmbeddingsFileURI(uri, None)


class EmbeddingsTextFile(Iterator[str]):
    """
    Utility class for opening embeddings text files. Handles various compression formats,
    as well as context management.

    # Parameters

    file_uri : `str`
        It can be:

        * a file system path or a URL of an eventually compressed text file or a zip/tar archive
          containing a single file.
        * URI of the type `(archive_path_or_url)#file_path_inside_archive` if the text file
          is contained in a multi-file archive.

    encoding : `str`
    cache_dir : `str`
    """

    DEFAULT_ENCODING = "utf-8"

    def __init__(
            self, file_uri: str, encoding: str = DEFAULT_ENCODING, cache_dir: str = None
    ) -> None:

        self.uri = file_uri
        self._encoding = encoding
        self._cache_dir = cache_dir
        self._archive_handle: Any = None  # only if the file is inside an archive

        main_file_uri, path_inside_archive = parse_embeddings_file_uri(file_uri)
        main_file_local_path = cached_path(main_file_uri, cache_dir=cache_dir)

        if zipfile.is_zipfile(main_file_local_path):  # ZIP archive
            self._open_inside_zip(main_file_uri, path_inside_archive)

        elif tarfile.is_tarfile(main_file_local_path):  # TAR archive
            self._open_inside_tar(main_file_uri, path_inside_archive)

        else:  # all the other supported formats, including uncompressed files
            if path_inside_archive:
                raise ValueError("Unsupported archive format: %s" + main_file_uri)

            # All the python packages for compressed files share the same interface of io.open
            extension = get_file_extension(main_file_uri)

            # Some systems don't have support for all of these libraries, so we import them only
            # when necessary.
            package = None
            if extension in [".txt", ".vec"]:
                package = io
            elif extension == ".gz":
                import gzip

                package = gzip
            elif extension == ".bz2":
                import bz2

                package = bz2
            elif extension == ".xz":
                import lzma

                package = lzma

            if package is None:
                logger.warning(
                    'The embeddings file has an unknown file extension "%s". '
                    "We will assume the file is an (uncompressed) text file",
                    extension,
                )
                package = io

            self._handle = package.open(  # type: ignore
                main_file_local_path, "rt", encoding=encoding
            )

        # To use this with tqdm we'd like to know the number of tokens. It's possible that the
        # first line of the embeddings file contains this: if it does, we want to start iteration
        # from the 2nd line, otherwise we want to start from the 1st.
        # Unfortunately, once we read the first line, we cannot move back the file iterator
        # because the underlying file may be "not seekable"; we use itertools.chain instead.
        first_line = next(self._handle)  # this moves the iterator forward
        self.num_tokens = EmbeddingsTextFile._get_num_tokens_from_first_line(first_line)
        if self.num_tokens:
            # the first line is a header line: start iterating from the 2nd line
            self._iterator = self._handle
        else:
            # the first line is not a header line: start iterating from the 1st line
            self._iterator = itertools.chain([first_line], self._handle)

    def _open_inside_zip(self, archive_path: str, member_path: Optional[str] = None) -> None:
        cached_archive_path = cached_path(archive_path, cache_dir=self._cache_dir)
        archive = zipfile.ZipFile(cached_archive_path, "r")
        if member_path is None:
            members_list = archive.namelist()
            member_path = self._get_the_only_file_in_the_archive(members_list, archive_path)
        member_path = cast(str, member_path)
        member_file = cast(BinaryIO, archive.open(member_path, "r"))
        self._handle = io.TextIOWrapper(member_file, encoding=self._encoding)
        self._archive_handle = archive

    def _open_inside_tar(self, archive_path: str, member_path: Optional[str] = None) -> None:
        cached_archive_path = cached_path(archive_path, cache_dir=self._cache_dir)
        archive = tarfile.open(cached_archive_path, "r")
        if member_path is None:
            members_list = archive.getnames()
            member_path = self._get_the_only_file_in_the_archive(members_list, archive_path)
        member_path = cast(str, member_path)
        member = archive.getmember(member_path)  # raises exception if not present
        member_file = cast(BinaryIO, archive.extractfile(member))
        self._handle = io.TextIOWrapper(member_file, encoding=self._encoding)
        self._archive_handle = archive

    def read(self) -> str:
        return "".join(self._iterator)

    def readline(self) -> str:
        return next(self._iterator)

    def close(self) -> None:
        self._handle.close()
        if self._archive_handle:
            self._archive_handle.close()

    def __enter__(self) -> "EmbeddingsTextFile":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def __iter__(self) -> "EmbeddingsTextFile":
        return self

    def __next__(self) -> str:
        return next(self._iterator)

    def __len__(self) -> Optional[int]:
        if self.num_tokens:
            return self.num_tokens
        raise AttributeError(
            "an object of type EmbeddingsTextFile implements `__len__` only if the underlying "
            "text file declares the number of tokens (i.e. the number of lines following)"
            "in the first line. That is not the case of this particular instance."
        )

    @staticmethod
    def _get_the_only_file_in_the_archive(members_list: Sequence[str], archive_path: str) -> str:
        if len(members_list) > 1:
            raise ValueError(
                "The archive %s contains multiple files, so you must select "
                "one of the files inside providing a uri of the type: %s."
                % (
                    archive_path,
                    format_embeddings_file_uri("path_or_url_to_archive", "path_inside_archive"),
                )
            )
        return members_list[0]

    @staticmethod
    def _get_num_tokens_from_first_line(line: str) -> Optional[int]:
        """This function takes in input a string and if it contains 1 or 2 integers, it assumes the
        largest one it the number of tokens. Returns None if the line doesn't match that pattern."""
        fields = line.split(" ")
        if 1 <= len(fields) <= 2:
            try:
                int_fields = [int(x) for x in fields]
            except ValueError:
                return None
            else:
                num_tokens = max(int_fields)
                logger.info(
                    "Recognized a header line in the embedding file with number of tokens: %d",
                    num_tokens,
                )
                return num_tokens
        return None



import functools
import inspect
import typing
from typing import Any, Callable, Dict, List, Optional

from combo.common.params import Params
from combo.config.exceptions import RegistryException
from combo.config.registry import (Registry)
from combo.utils import ConfigurationError


def _get_method_arguments(method: Callable) -> List[str]:
    return list(inspect.signature(method).parameters.keys())


def get_matching_arguments(args: Dict[str, Any], func: Callable) -> Dict[str, Any]:
    method_args: List[str] = _get_method_arguments(func)
    return {arg_name: arg_val for arg_name, arg_val in args if arg_name in method_args}


def _resolve(values: typing.Union[Dict[str, Any], str], pass_down_parameters: Dict[str, Any] = None) -> Any:
    if isinstance(values, Params):
        values = Params.as_dict()

    if isinstance(values, list):
        return [_resolve(v, pass_down_parameters) for v in values]

    if isinstance(values, dict):
        if 'type' not in values:
            return {vn: _resolve(vv, pass_down_parameters) for vn, vv in values.items()}

        try:
            clz, constructor = Registry.resolve(values['type'])
        except RegistryException:
            raise ConfigurationError('No registered class with name ' + str(values["type"]))

        args = _resolve(values.get('parameters', {}), pass_down_parameters)
        constructor_method = getattr(clz, constructor)
        if constructor_method is None:
            raise ConfigurationError(str(clz) + ' has no constructor method with name ' + constructor)

        return clz.from_parameters(args, constructor, pass_down_parameters)

    return values


def serialize_single_value(value: Any, pass_down_parameter_names: List[str] = None) -> Any:
    pass_down_parameter_names = pass_down_parameter_names or []
    if isinstance(value, FromParameters):
        return value.serialize(pass_down_parameter_names)
    elif isinstance(value, list):
        return [serialize_single_value(v, pass_down_parameter_names) for v in value]
    elif isinstance(value, tuple):
        return tuple([serialize_single_value(v, pass_down_parameter_names) for v in value])
    elif isinstance(value, set):
        return set([serialize_single_value(v, pass_down_parameter_names) for v in value])
    elif isinstance(value, dict):
        return {k: serialize_single_value(v, pass_down_parameter_names) for k, v in value.items()}
    elif isinstance(value, int) or isinstance(value, float) or isinstance(value, str):
        return value
    else:
        return str(value)


def register_arguments(func: callable):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self_arg = args[0]
        if self_arg.constructed_args is None:
            other_args = args[1:] if len(args) > 1 else []

            parameters_to_register = {}

            arg_names = _get_method_arguments(func)

            if len(arg_names) > 1:
                for arg_name, arg_val in zip(arg_names[1:], other_args):
                    parameters_to_register[arg_name] = arg_val
                for kwarg_name, kwarg_val in kwargs.items():
                    parameters_to_register[kwarg_name] = kwarg_val

            self_arg.constructed_args = parameters_to_register
        return func(*args, **kwargs)

    return wrapper


class FromParameters:
    constructed_from: str = "__init__"
    constructed_args: Optional[Dict[str, Any]] = None

    @classmethod
    def pass_down_parameter_names(cls) -> List[str]:
        return []

    @classmethod
    def from_parameters(cls,
                        parameters: Dict[str, Any] = None,
                        constructor_method_name: str = None,
                        pass_down_parameters: Dict[str, Any] = None):
        parameters_to_call = {}
        if parameters is None:
            parameters = {}

        if pass_down_parameters is None:
            pass_down_parameters = {}

        parameters = {**parameters, **pass_down_parameters}

        if constructor_method_name is None:
            constructor_method_name = '__init__'

        constructor_method = getattr(cls, constructor_method_name)
        if constructor_method is None:
            raise ValueError('No method ' + str(constructor_method) + ' for class ' + str(cls))

        constructor_method_args = _get_method_arguments(constructor_method)

        no_pass_down = {}

        # First, resolve the parameters that are needed to be passed down
        for param_name, param_value in parameters.items():
            if param_name in cls.pass_down_parameter_names():
                resolved_value = _resolve(param_value, pass_down_parameters)
                pass_down_parameters[param_name] = resolved_value

                if param_name in constructor_method_args:
                    parameters_to_call[param_name] = resolved_value
            else:
                no_pass_down[param_name] = param_value

        for param_name, param_value in no_pass_down.items():
            # Resolve only the arguments that are in the constructor method signature
            if param_name in constructor_method_args:
                parameters_to_call[param_name] = _resolve(param_value, pass_down_parameters)

        if constructor_method_name == '__init__':
            return cls(**parameters_to_call)
        else:
            return constructor_method(**parameters_to_call)

    def _to_params(self, pass_down_parameter_names: List[str] = None) -> Dict[str, str]:
        parameters_to_serialize = self.constructed_args or {}
        pass_down_parameter_names = pass_down_parameter_names or []

        parameters_dict = {}
        for pn, param_value in parameters_to_serialize.items():
            if pn in pass_down_parameter_names:
                continue
            parameters_dict[pn] = serialize_single_value(param_value,
                                                         pass_down_parameter_names + self.pass_down_parameter_names())
        return parameters_dict

    def serialize(self, pass_down_parameter_names: List[str] = None) -> Dict[str, Any]:
        constructor_method = self.constructed_from if self.constructed_from else '__init__'
        if not getattr(self, constructor_method):
            raise ConfigurationError('Class ' + str(type(self)) + ' has no constructor method ' + constructor_method)

        return {'type': Registry.get_class_name(type(self), constructor_method),
                'parameters': self._to_params(pass_down_parameter_names)}


def resolve(parameters: Dict[str, Any], pass_down_parameters: Dict[str, Any] = None) -> Any:
    pass_down_parameters = pass_down_parameters or {}
    clz, clz_init = Registry.resolve(parameters['type'])
    return clz.from_parameters(parameters['parameters'], clz_init, pass_down_parameters)


def flatten_dictionary(d, parent_key='', sep='/'):
    """
    Flatten a nested dictionary.

    Parameters:
        d (dict): The input dictionary.
        parent_key (str): The parent key to use for recursion (default is an empty string).
        sep (str): The separator to use when concatenating keys (default is '_').

    Returns:
        dict: A flattened dictionary.
    """
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dictionary(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def unflatten_dictionary(flat_dict, sep='/'):
    """
    Unflatten a flattened dictionary.

    Parameters:
        flat_dict (dict): The flattened dictionary.
        sep (str): The separator used in the flattened keys (default is '_').

    Returns:
        dict: The unflattened dictionary.
    """
    unflattened_dict = {}
    for key, value in flat_dict.items():
        keys = key.split(sep)
        current_level = unflattened_dict

        for k in keys[:-1]:
            current_level = current_level.setdefault(k, {})

        current_level[keys[-1]] = value

    return unflattened_dict


def override_parameters(parameters: Dict[str, Any], override_values: Dict[str, Any]) -> Dict[str, Any]:
    overriden_parameters = flatten_dictionary(parameters)
    override_values = flatten_dictionary(override_values)
    for ko, vo in override_values.items():
        #if ko in overriden_parameters:
        overriden_parameters[ko] = vo

    return unflatten_dictionary(overriden_parameters)


def override_or_add_parameters(parameters: Dict[str, Any], override_values: Dict[str, Any]) -> Dict[str, Any]:
    overriden_parameters = flatten_dictionary(parameters)
    override_values = flatten_dictionary(override_values)
    for ko, vo in override_values.items():
        overriden_parameters[ko] = vo

    return unflatten_dictionary(overriden_parameters)



import os
import shutil
import tempfile
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Union, NamedTuple, Optional

import cached_path
import torch
import json
import tarfile
from io import BytesIO
from tempfile import TemporaryDirectory

from combo.config import resolve
from combo.data.dataset_loaders import DataLoader
from combo.data.dataset_readers import DatasetReader
from combo.modules.model import Model
from contextlib import contextmanager

import logging
from combo.utils import ComboLogger

logging.setLoggerClass(ComboLogger)
logger = logging.getLogger(__name__)


CACHE_ROOT = Path(os.getenv("COMBO_CACHE_ROOT", Path.home() / ".combo"))
CACHE_DIRECTORY = str(CACHE_ROOT / "cache")

PREFIX = 'Loading archive'

class Archive(NamedTuple):
    model: Model
    config: Optional[Dict[str, Any]]
    data_loader: Optional[DataLoader]
    validation_data_loader: Optional[DataLoader]
    dataset_reader: Optional[DatasetReader]


def add_to_tar(tar_file: tarfile.TarFile, out_stream: BytesIO, data: bytes, name: str):
    out_stream.write(data)
    out_stream.seek(0)
    info = tarfile.TarInfo(name)
    info.size = len(out_stream.getbuffer())
    tar_file.addfile(info, out_stream)


def archive(model: Model,
            serialization_dir: Union[PathLike, str],
            data_loader: Optional[DataLoader] = None,
            validation_data_loader: Optional[DataLoader] = None,
            dataset_reader: Optional[DatasetReader] = None) -> str:
    parameters = {'vocabulary': {
        'type': 'from_files_vocabulary',
        'parameters': {
            'directory': 'vocabulary',
            'padding_token': model.vocab._padding_token,
            'oov_token': model.vocab._oov_token
        }
    }, 'model': model.serialize(pass_down_parameter_names=['vocabulary', 'optimizer', 'scheduler', 'model_name'])}

    if data_loader:
        parameters['data_loader'] = data_loader.serialize()
    if validation_data_loader:
        parameters['validation_data_loader'] = validation_data_loader.serialize()
    if dataset_reader:
        parameters['dataset_reader'] = dataset_reader.serialize()

    parameters['training'] = {}

    if model.optimizer:
        parameters['training']['optimizer'] = model.optimizer.serialize()

    if model.scheduler:
        parameters['training']['scheduler'] = model.scheduler.serialize(pass_down_parameter_names=['optimizer'])

    if model.model_name:
        parameters['model_name'] = model.model_name

    with (TemporaryDirectory(os.path.join('tmp')) as t,
          BytesIO() as out_stream,
          tarfile.open(os.path.join(serialization_dir, 'model.tar.gz'), 'w|gz') as tar_file):
        add_to_tar(tar_file, out_stream, json.dumps(parameters).encode(), 'config.json')
        weights_path = os.path.join(t, 'weights.th')
        torch.save(model.state_dict(), weights_path)
        tar_file.add(weights_path, 'weights.th')
        vocabulary_path = os.path.join(t, 'vocabulary')
        model.vocab.save_to_files(vocabulary_path)
        tar_file.add(vocabulary_path, 'vocabulary', recursive=True)

    return serialization_dir


@contextmanager
def extracted_archive(resolved_archive_file, cleanup=True):
    tempdir = None
    try:
        tempdir = tempfile.mkdtemp(dir=CACHE_DIRECTORY)
        with tarfile.open(resolved_archive_file) as archive:
            subdir_and_files = [
                tarinfo for tarinfo in archive.getmembers()
                if (any([tarinfo.name.endswith(f) for f in ['config.json', 'weights.th']])
                    or 'vocabulary' in tarinfo.name)
            ]
            for f in subdir_and_files:
                if 'vocabulary' in f.name and not f.name.endswith('vocabulary'):
                    f.name = os.path.join('vocabulary', os.path.basename(f.name))
                else:
                    f.name = os.path.basename(f.name)
            archive.extractall(path=tempdir, members=subdir_and_files)
            yield tempdir
    finally:
        if tempdir is not None and cleanup:
            shutil.rmtree(tempdir, ignore_errors=True)


def load_archive(url_or_filename: Union[PathLike, str],
                 cache_dir: Union[PathLike, str] = None,
                 cuda_device: int = -1) -> Archive:

    rarchive_file = cached_path.cached_path(
            url_or_filename,
            cache_dir=cache_dir or CACHE_DIRECTORY,
        )

    with extracted_archive(rarchive_file) as archive_file:
        model = Model.load(archive_file, cuda_device=cuda_device)

        config_path = os.path.join(archive_file, 'config.json')
        with open(config_path, 'r') as f:
            config = json.load(f)

        data_loader, validation_data_loader, dataset_reader = None, None, None
        pass_down_parameters = {}
        if config.get("model_name"):
            pass_down_parameters = {"model_name": config.get("model_name")}


        if 'data_loader' in config:
            try:
                data_loader = resolve(config['data_loader'],
                                      pass_down_parameters=pass_down_parameters)
            except Exception as e:
                logger.warning(f'Error while loading Training Data Loader: {str(e)}. Setting Data Loader to None',
                               prefix=PREFIX)
        if 'validation_data_loader' in config:
            try:
                validation_data_loader = resolve(config['validation_data_loader'],
                                                 pass_down_parameters=pass_down_parameters)
            except Exception as e:
                logger.warning(f'Error while loading Validation Data Loader: {str(e)}. Setting Data Loader to None',
                               prefix=PREFIX)
        if 'dataset_reader' in config:
            try:
                dataset_reader = resolve(config['dataset_reader'],
                                         pass_down_parameters=pass_down_parameters)
            except Exception as e:
                logger.warning(f'Error while loading Dataset Reader: {str(e)}. Setting Dataset Reader to None',
                               prefix=PREFIX)

    return Archive(model=model,
                   config=config,
                   data_loader=data_loader,
                   validation_data_loader=validation_data_loader,
                   dataset_reader=dataset_reader)



"""Based on https://github.com/emorynlp/iwpt-shared-task-2020."""

import numpy as np

_ACL_REL_CL = "acl:relcl"


def graph_and_tree_merge(tree_arc_scores,
                         tree_rel_scores,
                         graph_arc_scores,
                         graph_rel_scores,
                         label2idx,
                         idx2label,
                         graph_label2idx,
                         graph_idx2label,
                         tokens):
    graph_arc_scores = np.copy(graph_arc_scores)
    # Exclude self-loops, in-place operation.
    np.fill_diagonal(graph_arc_scores, 0)
    # Connection to root will be handled by tree.
    graph_arc_scores[:, 0] = False
    # The same with labels.
    root_idx = graph_label2idx["root"]
    graph_rel_scores[:, :, root_idx] = -float('inf')
    graph_rel_pred = graph_rel_scores.argmax(-1)

    # Add tree edges to graph
    tree_heads = [0] + tree_arc_scores
    graph = [[] for _ in range(len(tree_heads))]
    labeled_graph = [[] for _ in range(len(tree_heads))]
    for d, h in enumerate(tree_heads):
        if not d:
            continue
        label = idx2label[tree_rel_scores[d - 1]]
        # graph_label = graph_idx2label[graph_rel_pred[d - 1][h - 1]]
        # if ">" in graph_label and label in graph_label:
        #     print("Using graph label instead of tree.")
        #     label = graph_label
        if label != _ACL_REL_CL:
            graph[h].append(d)
            labeled_graph[h].append((d, label))

    # Debug only
    # Extract graph edges
    graph_edges = np.argwhere(graph_arc_scores)

    # Add graph edges which aren't creating a cycle
    for (d, h) in graph_edges:
        if not d or not h or d in graph[h]:
            continue
        try:
            path = next(_dfs(graph, d, h))
        except StopIteration:
            # There is not path from d to h
            label = graph_idx2label[graph_rel_pred[d][h]]
            if label != _ACL_REL_CL:
                graph[h].append(d)
                labeled_graph[h].append((d, label))

    # Add 'acl:relcl' without checking for cycles.
    for d, h in enumerate(tree_heads):
        if not d:
            continue
        label = idx2label[tree_rel_scores[d - 1]]
        if label == _ACL_REL_CL:
            graph[h].append(d)
            labeled_graph[h].append((d, label))

    assert len(labeled_graph[0]) == 1
    d = graph[0][0]
    graph[d].append(0)
    labeled_graph[d].append((0, "root"))

    parse_graph = [[] for _ in range(len(tree_heads))]
    for h in range(len(tree_heads)):
        for d, label in labeled_graph[h]:
            parse_graph[d].append((h, label))
        parse_graph[d] = sorted(parse_graph[d])

    for i, g in enumerate(parse_graph):
        heads = np.array([x[0] for x in g])
        rels = np.array([x[1] for x in g])
        indices = rels.argsort()
        heads = heads[indices].tolist()
        rels = rels[indices].tolist()
        deps = '|'.join(f'{h}:{r}' for h, r in zip(heads, rels))
        tokens[i - 1]["deps"] = deps
    return


def _dfs(graph, start, end):
    fringe = [(start, [])]
    while fringe:
        state, path = fringe.pop()
        if path and state == end:
            yield path
            continue
        for next_state in graph[state]:
            if next_state in path:
                continue
            fringe.append((next_state, path + [next_state]))


def restore_collapse_edges(tree_tokens):
    # https://gist.github.com/hankcs/776e7d95c19e5ff5da8469fe4e9ab050
    empty_tokens = []
    for token in tree_tokens:
        deps = token["deps"].split("|")
        for i, d in enumerate(deps):
            if ">" in d:
                # {head}:{empty_node_relation}>{current_node_relation}
                # should map to
                # For new, empty node:
                # {head}:{empty_node_relation}
                # For current node:
                # {new_empty_node_id}:{current_node_relation}
                # TODO consider where to put new_empty_node_id (currently at the end)
                head, relation = d.split(':', 1)
                ehead = f"{len(tree_tokens)}.{len(empty_tokens) + 1}"
                empty_node_relation, current_node_relation = relation.split(">", 1)
                # Edge case, double >
                if ">" in current_node_relation:
                    second_empty_node_relation, current_node_relation = current_node_relation.split(">")
                    deps[i] = f"{ehead}:{current_node_relation}"
                    second_ehead = f"{len(tree_tokens)}.{len(empty_tokens) + 2}"
                    empty_tokens.append(
                        {
                            "idx": ehead,
                            "deps": f"{second_ehead}:{empty_node_relation}"
                        }
                    )
                    empty_tokens.append(
                        {
                            "idx": second_ehead,
                            "deps": f"{head}:{second_empty_node_relation}"
                        }
                    )

                else:
                    deps[i] = f"{ehead}:{current_node_relation}"
                    empty_tokens.append(
                        {
                            "idx": ehead,
                            "deps": f"{head}:{empty_node_relation}"
                        }
                    )
        deps = sorted([d.split(":", 1) for d in deps], key=lambda x: float(x[0]))
        token["deps"] = "|".join([f"{k}:{v}" for k, v in deps])
    return empty_tokens



from typing import Any, Dict, List, Optional, Type

import pytorch_lightning as pl
import torch
from torch import Tensor

from combo.config import FromParameters
from combo.data.dataset_loaders.dataset_loader import TensorDict
from combo.modules.model import Model
from combo.training import Scheduler


class TrainableCombo(pl.LightningModule, FromParameters):
    def __init__(self,
                 model: Model,
                 optimizer_type: Type = torch.optim.Adam,
                 optimizer_kwargs: Optional[Dict[str, Any]] = None,
                 scheduler_type: Type = Scheduler,
                 scheduler_kwargs: Optional[Dict[str, Any]] = None,
                 validation_metrics: List[str] = None):
        super().__init__()
        self.model = model
        self._optimizer_type = optimizer_type
        self._optimizer_kwargs = optimizer_kwargs if optimizer_kwargs else {}

        self._scheduler_type = scheduler_type
        self._scheduler_kwargs = scheduler_kwargs if scheduler_kwargs else {}

        self._validation_metrics = validation_metrics if validation_metrics else []

    def forward(self, batch: TensorDict) -> TensorDict:
        return self.model.batch_outputs(batch, self.model.training)

    def training_step(self, batch: TensorDict, batch_idx: int) -> Tensor:
        output = self.forward(batch)
        self.log("train_loss", output['loss'], on_step=True, on_epoch=True, prog_bar=True, logger=True)
        return output["loss"]

    def validation_step(self, batch: TensorDict, batch_idx: int) -> Tensor:
        output = self.forward(batch)
        metrics = self.model.get_metrics()
        for k in metrics.keys():
            if k in self._validation_metrics:
                self.log(k, metrics[k], on_epoch=True, prog_bar=True, logger=True)
        return output["loss"]

    def lr_scheduler_step(self, scheduler: torch.optim.lr_scheduler, metric: Optional[Any]) -> None:
        scheduler.step(metric=metric)

    def configure_optimizers(self):
        optimizer = self._optimizer_type(self.model.parameters(), **self._optimizer_kwargs)
        return ([optimizer],
                [{'scheduler': self._scheduler_type(optimizer, **self._scheduler_kwargs),
                  'interval': 'epoch'}])



import torch

from combo.config import Registry
from combo.config.from_parameters import register_arguments, FromParameters


@Registry.register('base_scheduler')
class Scheduler(torch.optim.lr_scheduler.LambdaLR, FromParameters):
    pass_down_parameters = ['optimizer']

    @register_arguments
    def __init__(self,
                 optimizer: torch.optim.Optimizer,
                 patience: int = 6,
                 decreases: int = 2,
                 threshold: float = 1e-3,
                 last_epoch: int = -1,
                 verbose: bool = False):
        super().__init__(optimizer, [self._lr_lambda], last_epoch, verbose)
        self.patience = patience
        self.decreases = decreases
        self.threshold = threshold
        self.start_patience = patience
        self.best_score = 0.0

    @staticmethod
    def _lr_lambda(idx: int) -> float:
        return 1.0 / (1.0 + idx * 1e-4)

    def step(self, metric: float = None) -> None:
        super().step()

        if metric is not None:
            if metric - self.best_score > self.threshold:
                self.best_score = metric if metric > self.best_score else self.best_score
                self.patience = self.start_patience
            else:
                if self.patience <= 1:
                    if self.decreases == 0:
                        # The Trainer should trigger early stopping
                        self.patience = 0
                    else:
                        self.patience = self.start_patience
                        self.decreases -= 1
                        self.threshold /= 2
                        self.base_lrs = [x / 2 for x in self.base_lrs]
                else:
                    self.patience -= 1
